#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import os
import re
import enum
import base64
import asyncio
import datetime
import threading
import traceback
import multiprocessing

from jackdaw.dbmodel.spnservice import JackDawSPNService
from jackdaw.dbmodel.addacl import JackDawADDACL
from jackdaw.dbmodel.adgroup import JackDawADGroup
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.aduser import JackDawADUser
from jackdaw.dbmodel.adcomp import JackDawADMachine
from jackdaw.dbmodel.adou import JackDawADOU
from jackdaw.dbmodel.usergroup import JackDawGroupUser
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.tokengroup import JackDawTokenGroup
from jackdaw.dbmodel.adgpo import JackDawADGPO
from jackdaw.dbmodel.constrained import JackDawMachineConstrainedDelegation, JackDawUserConstrainedDelegation
from jackdaw.dbmodel.adgplink import JackDawADGplink
from jackdaw.dbmodel.adsd import JackDawSD
from jackdaw.dbmodel.adtrust import JackDawADTrust
from jackdaw.dbmodel import get_session
from jackdaw.wintypes.lookup_tables import *
from jackdaw import logger

from jackdaw.common.apq import AsyncProcessQueue

from msldap.ldap_objects import *
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR
from tqdm import tqdm

class LDAPEnumeratorProgress:
	def __init__(self):
		self.type = 'LDAP'
		self.msg_type = 'PROGRESS'
		self.adid = None
		self.domain_name = None
		self.finished = None
		self.running = None
		self.total_finished = None
		self.speed = None #per sec

	def __str__(self):
		if self.msg_type == 'PROGRESS':
			return '[%s][%s][%s][%s] FINISHED %s RUNNING %s TOTAL %s SPEED %s' % (
				self.type, 
				self.domain_name, 
				self.adid,
				self.msg_type,
				','.join(self.finished), 
				','.join(self.running), 
				self.total_finished, 
				self.speed
			
			)
		return '[%s][%s][%s][%s]' % (self.type, self.domain_name, self.adid, self.msg_type)

class LDAPAgentCommand(enum.Enum):
	SPNSERVICE = 0
	SPNSERVICES = 1
	USER = 2
	USERS = 3
	MACHINE = 4
	MACHINES = 5
	OU = 6
	OUS = 7
	DOMAININFO = 8
	GROUP = 9
	GROUPS = 10
	MEMBERSHIP = 11
	MEMBERSHIPS = 12
	SD = 13
	SDS = 14
	GPO = 15
	GPOS = 16
	TRUSTS = 17
	EXCEPTION = 99

	SPNSERVICES_FINISHED = 31
	USERS_FINISHED = 32
	MACHINES_FINISHED = 33
	OUS_FINISHED = 34
	GROUPS_FINISHED = 35
	MEMBERSHIPS_FINISHED = 36
	SDS_FINISHED = 37
	DOMAININFO_FINISHED = 38
	GPOS_FINISHED = 39
	TRUSTS_FINISHED = 40

MSLDAP_JOB_TYPES = {
	'users' : LDAPAgentCommand.USERS_FINISHED ,
	'machines' : LDAPAgentCommand.MACHINES_FINISHED ,
	'sds' : LDAPAgentCommand.SDS_FINISHED ,
	'memberships' : LDAPAgentCommand.MEMBERSHIPS_FINISHED ,
	'ous' : LDAPAgentCommand.OUS_FINISHED ,
	'gpos' : LDAPAgentCommand.GPOS_FINISHED ,
	'groups' : LDAPAgentCommand.GROUPS_FINISHED ,
	'spns' : LDAPAgentCommand.SPNSERVICES_FINISHED ,
	'adinfo' : LDAPAgentCommand.DOMAININFO_FINISHED,
	'trusts' : LDAPAgentCommand.TRUSTS_FINISHED
}
MSLDAP_JOB_TYPES_INV = {v: k for k, v in MSLDAP_JOB_TYPES.items()}

class LDAPAgentJob:
	def __init__(self, command, data):
		self.command = command
		self.data = data

class LDAPEnumeratorAgent():
	def __init__(self, ldap_mgr, agent_in_q, agent_out_q):
		#multiprocessing.Process.__init__(self)
		self.ldap_mgr = ldap_mgr
		self.agent_in_q = agent_in_q
		self.agent_out_q = agent_out_q
		self.ldap = None
		self.test_ctr = 0

	async def get_all_effective_memberships(self):
		try:
			async for res in self.ldap.get_all_tokengroups():
				s = JackDawTokenGroup()
				s.cn = res['cn']
				s.dn = res['dn']
				s.guid = res['guid']
				s.sid = res['sid']
				s.member_sid = res['token']
				s.is_user = True if res['type'] == 'user' else False
				s.is_group = True if res['type'] == 'group' else False
				s.is_machine = True if res['type'] == 'computer' else False
				await self.agent_out_q.put((LDAPAgentCommand.MEMBERSHIP, s))
		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			await self.agent_out_q.put((LDAPAgentCommand.MEMBERSHIPS_FINISHED, None))

	async def enumerate_spnservices(self):
		pass

	async def enumerate_machine(self, machine):
		pass

	async def get_all_trusts(self):
		try:
			async for entry in self.ldap.get_all_trusts():
				await self.agent_out_q.put((LDAPAgentCommand.TRUSTS, JackDawADTrust.from_ldapdict(entry.to_dict())))
		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			await self.agent_out_q.put((LDAPAgentCommand.TRUSTS_FINISHED, None))

	async def get_all_spnservices(self):
		try:
			async for entry in self.ldap.get_all_spn_entries():
				for spn in entry['attributes']['servicePrincipalName']:
					port = None
					service, t = spn.rsplit('/',1)
					m = t.find(':')
					if m != -1:
						computername, port = spn.rsplit(':',1)
					else:
						computername = t

					s = JackDawSPNService()
					s.owner_sid = str(entry['attributes']['objectSid'])
					s.computername = computername
					s.service = service
					s.port = port
					await self.agent_out_q.put((LDAPAgentCommand.SPNSERVICE, s))
		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			await self.agent_out_q.put((LDAPAgentCommand.SPNSERVICES_FINISHED, None))

	async def get_all_users(self):
		try:
			async for user_data in self.ldap.get_all_user_objects():
				user = JackDawADUser.from_aduser(user_data)
				await self.agent_out_q.put((LDAPAgentCommand.USER, user))
		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			await self.agent_out_q.put((LDAPAgentCommand.USERS_FINISHED, None))

	async def get_all_groups(self):
		try:
			async for group in self.ldap.get_all_groups():
				g = JackDawADGroup.from_dict(group.to_dict())
				await self.agent_out_q.put((LDAPAgentCommand.GROUP, g))
				del g
		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			await self.agent_out_q.put((LDAPAgentCommand.GROUPS_FINISHED, None))

	async def get_all_gpos(self):
		try:
			async for gpo in self.ldap.get_all_gpos():
				g = JackDawADGPO.from_adgpo(gpo)
				await self.agent_out_q.put((LDAPAgentCommand.GPO, g))
				del g
		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			await self.agent_out_q.put((LDAPAgentCommand.GPOS_FINISHED, None))


	async def get_all_machines(self):
		try:
			async for machine_data in self.ldap.get_all_machine_objects():
				machine = JackDawADMachine.from_adcomp(machine_data)
				await self.agent_out_q.put((LDAPAgentCommand.MACHINE, machine))
		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			await self.agent_out_q.put((LDAPAgentCommand.MACHINES_FINISHED, None))

	async def get_all_ous(self):
		try:
			async for ou in self.ldap.get_all_ous():
				o = JackDawADOU.from_adou(ou)
				await self.agent_out_q.put((LDAPAgentCommand.OU, o))
				del o
		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			await self.agent_out_q.put((LDAPAgentCommand.OUS_FINISHED, None))

	async def get_domain_info(self):
		try:
			info = await self.ldap.get_ad_info()
			adinfo = JackDawADInfo.from_dict(info.to_dict())
			await self.agent_out_q.put((LDAPAgentCommand.DOMAININFO, adinfo))
		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			await self.agent_out_q.put((LDAPAgentCommand.DOMAININFO_FINISHED, None))

	async def get_sds(self, data):
		try:
			async for adsec in self.ldap.get_all_objectacl():
				if not adsec.nTSecurityDescriptor:
					continue
				await self.agent_out_q.put((LDAPAgentCommand.SD, adsec ))

		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
		finally:
			await self.agent_out_q.put((LDAPAgentCommand.SDS_FINISHED, None))

	async def setup(self):
		try:
			self.ldap = self.ldap_mgr.get_client()
			res, err = await self.ldap.connect()
			if err is not None:
				raise err
			return res
		except:
			await self.agent_out_q.put((LDAPAgentCommand.EXCEPTION, str(traceback.format_exc())))
			return False

	async def arun(self):
		res = await self.setup()
		if res is False:
			return
		while True:
			res = await self.agent_in_q.get()
			if res is None:
				return

			if res.command == LDAPAgentCommand.DOMAININFO:
				await self.get_domain_info()
			elif res.command == LDAPAgentCommand.USERS:
				await self.get_all_users()
			elif res.command == LDAPAgentCommand.MACHINES:
				await self.get_all_machines()
			elif res.command == LDAPAgentCommand.GROUPS:
				await self.get_all_groups()
			elif res.command == LDAPAgentCommand.OUS:
				await self.get_all_ous()
			elif res.command == LDAPAgentCommand.GPOS:
				await self.get_all_gpos()
			elif res.command == LDAPAgentCommand.SPNSERVICES:
				await self.get_all_spnservices()
			elif res.command == LDAPAgentCommand.MEMBERSHIPS:
				await self.get_all_effective_memberships()
			elif res.command == LDAPAgentCommand.SDS:
				await self.get_sds(res.data)
			elif res.command == LDAPAgentCommand.TRUSTS:
				await self.get_all_trusts()

	def run(self):
		try:
			loop = asyncio.get_event_loop()
		except:
			loop = asyncio.new_event_loop()
		#loop.set_debug(True)  # Enable debug
		loop.run_until_complete(self.arun())

class LDAPEnumeratorManager:
	def __init__(self, db_conn, ldam_mgr, agent_cnt = None, queue_size = None, progress_queue = None):
		self.db_conn = db_conn
		self.ldam_mgr = ldam_mgr

		self.session = None

		self.queue_size = queue_size
		self.agent_in_q = None
		self.agent_out_q = None
		self.agents = []

		self.agent_cnt = agent_cnt
		if agent_cnt is None:
			self.agent_cnt = min(multiprocessing.cpu_count(), 3)

		self.ad_id = None
		self.domain_name = None

		self.user_ctr = 0
		self.machine_ctr = 0
		self.ou_ctr = 0
		self.group_ctr = 0
		self.sd_ctr = 0
		self.spn_ctr = 0
		self.member_ctr = 0
		self.domaininfo_ctr = 0
		self.gpo_ctr = 0
		self.trust_ctr = 0

		self.user_finish_ctr = 0
		self.machine_finish_ctr = 0
		self.ou_finish_ctr = 0
		self.group_finish_ctr = 0
		self.sd_finish_ctr = 0
		self.spn_finish_ctr = 0
		self.member_finish_ctr = 0
		self.domaininfo_finish_ctr = 0
		self.gpo_finish_ctr = 0

		self.progress_queue = progress_queue
		self.total_progress = None
		self.total_counter = 0
		self.total_counter_steps = 100
		self.progress_total_present = False
		self.remaining_ctr = None
		self.progress_last_updated = datetime.datetime.utcnow()
		self.progress_last_counter = 0

		self.enum_finished_evt = None #multiprocessing.Event()
		
		self.running_enums = {}
		self.finished_enums = []
		self.enum_types = [
			'adinfo',
			'trusts',
			'users', 
			'machines',
			'groups',
			'memberships', 
			'sds', 
			'ous', 
			'gpos',
			'spns'
		]
		self.enum_types_len = len(self.enum_types)

	async def check_jobs(self, finished_type):
		self.session.commit()
		if finished_type is not None:
			logger.debug('%s enumeration finished!' % MSLDAP_JOB_TYPES_INV[finished_type])
			del self.running_enums[MSLDAP_JOB_TYPES_INV[finished_type]]
			self.finished_enums.append(MSLDAP_JOB_TYPES_INV[finished_type])

		lr = len(self.running_enums)
		if self.enum_types_len == len(self.finished_enums):
			#everything finished
			return True

		if lr == self.agent_cnt:
			#enums still running with max performance
			return False

		if lr < self.agent_cnt:
			#we can start a new enum
			for _ in range(self.agent_cnt - lr):
				if len(self.enum_types) > 0:
					next_type = self.enum_types.pop(0)
				else:
					return False
				
				if next_type == 'adinfo':
					await self.enum_domain()
					#this must be the first!
					self.running_enums[next_type] = 1
					return False
				
				elif next_type == 'users':
					await self.enum_users()

				elif next_type == 'machines':
					await self.enum_machines()
				elif next_type == 'sds':
					await self.enum_sds()
				elif next_type == 'memberships':
					await self.enum_memberships()
				elif next_type == 'ous':
					await self.enum_ous()
				elif next_type == 'gpos':
					await self.enum_gpos()
				elif next_type == 'groups':
					await self.enum_groups()
				elif next_type == 'spns':
					await self.enum_spnservices()
				elif next_type == 'trusts':
					await self.enum_trusts()
				else:
					logger.warning('Unknown next_type! %s' % next_type)

				self.running_enums[next_type] = 1

			return False
		

	@staticmethod
	def spn_to_account(spn):
		if spn.find('/') != -1:
			return spn.rsplit('/')[1].upper() + '$'
	
	def get_enum_stats(self):
		return {
			'users' : self.user_ctr,
			'machines' : self.machine_ctr,
			'ous' : self.ou_ctr,
			'groups' : self.group_ctr,
			'security_descriptors' : self.sd_ctr,
			'spns' : self.spn_ctr,
			'membership_tokens' : self.member_ctr,
			'domaininfo' : self.domaininfo_ctr,
			'gpos' : self.gpo_ctr,
		}

	async def setup(self):
		logger.debug('mgr setup')

		qs = self.queue_size
		if qs is None:
			qs = self.agent_cnt
		self.agent_in_q = asyncio.Queue() #AsyncProcessQueue()
		self.agent_out_q = asyncio.Queue(qs) #AsyncProcessQueue(1000)

		if self.progress_queue is None:
			self.total_progress = tqdm(desc='LDAP info entries', ascii = True)
		
		self.session = get_session(self.db_conn)
		
		for _ in range(self.agent_cnt):
			agent = LDAPEnumeratorAgent(self.ldam_mgr, self.agent_in_q, self.agent_out_q)
			self.agents.append(asyncio.create_task(agent.arun()))
			#agent.daemon = True
			#agent.start()
			#self.agents.append(agent)

	async def terminate(self):
		logger.debug('terminate called!')
		if self.progress_queue is None:
			msg = LDAPEnumeratorProgress()
			msg.type = 'LDAP'
			msg.msg_type = 'ABORTED'
			msg.adid = self.ad_id
			msg.domain_name = self.domain_name
			await self.progress_queue.put(msg)

		for task in self.agents:
			task.cancel()
		try:
			self.session.commit()
		except:
			pass
		
		try:
			self.session.close()
		except:
			pass

	async def enum_domain(self):
		logger.debug('Enumerating domain')
		job = LDAPAgentJob(LDAPAgentCommand.DOMAININFO, None)
		await self.agent_in_q.put(job)

	async def store_domain(self, info):
		info.ldap_enumeration_state = 'STARTED'
		self.domain_name = str(info.distinguishedName).replace(',','.').replace('DC=','')
		self.session.add(info)
		self.session.commit()
		self.session.refresh(info)
		self.ad_id = info.id

	async def enum_trusts(self):
		logger.debug('Enumerating trusts')
		job = LDAPAgentJob(LDAPAgentCommand.TRUSTS, None)
		await self.agent_in_q.put(job)

	async def store_trust(self, trust):
		trust.ad_id = self.ad_id
		self.session.add(trust)
		self.session.flush()

	async def enum_users(self):
		logger.debug('Enumerating users')
		job = LDAPAgentJob(LDAPAgentCommand.USERS, self.ad_id)
		await self.agent_in_q.put(job)
		

	async def store_user(self, user):
		user.ad_id = self.ad_id

		#self.session.flush()
		#self.session.refresh(user)

		for spn in getattr(user,'allowedtodelegateto',[]):
			con = JackDawUserConstrainedDelegation()
			con.spn = spn
			con.targetaccount = LDAPEnumeratorManager.spn_to_account(spn)
			user.allowedtodelegateto.append(con)

		self.session.add(user)
		self.session.flush()

	async def enum_machines(self):
		logger.debug('Enumerating machines')
		job = LDAPAgentJob(LDAPAgentCommand.MACHINES, self.ad_id)
		await self.agent_in_q.put(job)

	async def store_machine(self, machine):
		machine.ad_id = self.ad_id
		self.session.add(machine)
		#self.session.commit()
		self.session.flush()
	
	async def enum_groups(self):
		logger.debug('Enumerating groups')
		job = LDAPAgentJob(LDAPAgentCommand.GROUPS, self.ad_id)
		await self.agent_in_q.put(job)

	async def store_group(self, group):
		group.ad_id = self.ad_id
		self.session.add(group)
		self.session.flush()

	async def enum_ous(self):
		logger.debug('Enumerating ous')
		job = LDAPAgentJob(LDAPAgentCommand.OUS, self.ad_id)
		await self.agent_in_q.put(job)

	async def store_ous(self, ou):
		ou.ad_id = self.ad_id
		self.session.add(ou)
		self.session.commit()
		self.session.refresh(ou)

		if ou.gPLink is not None and ou.gPLink != 'None':
			for x in ou.gPLink.split(']'):
				if x is None or x == 'None':
					continue
				x = x.strip()
				if x == '':
					continue
				gp, order = x[1:].split(';')
				gp = re.search(r'{(.*?)}', gp).group(1)
				gp = '{' + gp + '}'

				link = JackDawADGplink()
				link.ent_id = ou.id
				link.gpo_dn = gp
				link.order = order
				self.session.add(link)
		self.session.flush()

	async def enum_spnservices(self):		
		logger.debug('Enumerating spns')
		job = LDAPAgentJob(LDAPAgentCommand.SPNSERVICES, self.ad_id)
		await self.agent_in_q.put(job)

	async def store_spn(self, spn):
		spn.ad_id = self.ad_id
		self.session.add(spn)
		self.session.flush()

	async def enum_gpos(self):
		logger.debug('Enumerating gpos')
		job = LDAPAgentJob(LDAPAgentCommand.GPOS, self.ad_id)
		await self.agent_in_q.put(job)

	async def store_gpo(self, gpo):
		gpo.ad_id = self.ad_id
		self.session.add(gpo)
		self.session.flush()

	async def enum_memberships(self):
		logger.debug('Enumerating memberships')
		job = LDAPAgentJob(LDAPAgentCommand.MEMBERSHIPS, None)
		await self.agent_in_q.put(job)


	async def store_membership(self, res):
		res.ad_id = self.ad_id
		self.session.add(res)
		
		if self.member_finish_ctr % 1000 == 0:
			self.session.commit()
		else:
			self.session.flush()

	async def enum_sds(self):
		logger.debug('Enumerating security descriptors')
		job = LDAPAgentJob(LDAPAgentCommand.SDS, None)
		await self.agent_in_q.put(job)

	async def store_sd(self, sd):
		#secdesc = SECURITY_DESCRIPTOR.from_bytes(sd.nTSecurityDescriptor)
		#
		if sd.objectClass[-1] in ['user', 'group']:
			obj_type = sd.objectClass[-1]
		elif sd.objectClass[-1] == 'computer':
			obj_type = 'machine'
		elif sd.objectClass[-1] == 'groupPolicyContainer':
			obj_type = 'gpo'
		elif sd.objectClass[-1] == 'organizationalUnit':
			obj_type = 'ou'
		else:
			obj_type = sd.objectClass[-1]

		jdsd = JackDawSD()

		jdsd.ad_id = self.ad_id
		jdsd.guid =  str(sd.objectGUID)
		if sd.objectSid:
			jdsd.sid = str(sd.objectSid)
		jdsd.object_type = obj_type
		jdsd.sd = base64.b64encode(sd.nTSecurityDescriptor)
		
		self.session.add(jdsd)
		
		if self.sd_ctr % 1000 == 0:
			self.session.commit()
		else:
			self.session.flush()

	async def update_progress(self):
		self.total_counter += 1

		if self.progress_queue is None:
			#if self.remaining_ctr is not None and self.progress_total_present is False:
			#	self.total_progress.total = (self.remaining_ctr +  self.total_counter)
			#	self.progress_total_present = True
			
			if self.total_counter % self.total_counter_steps == 0:
				self.total_progress.update(self.total_counter_steps)

			if self.total_counter % 5000 == 0:
				running_jobs = ','.join([k for k in self.running_enums])
				finished_jobs = ','.join(self.finished_enums)
				msg = 'FINISHED: %s RUNNING: %s' % (finished_jobs, running_jobs)
				#logger.debug(msg)
				self.total_progress.set_description(msg)
				self.total_progress.refresh()
		
		else:
			if self.total_counter % self.total_counter_steps == 0:
				now = datetime.datetime.utcnow()
				td = (now - self.progress_last_updated).total_seconds()
				self.progress_last_updated = now
				cd = self.total_counter - self.progress_last_counter
				self.progress_last_counter = self.total_counter
				msg = LDAPEnumeratorProgress()
				msg.type = 'LDAP'
				msg.msg_type = 'PROGRESS'
				msg.adid = self.ad_id
				msg.domain_name = self.domain_name
				msg.finished = self.finished_enums
				msg.running = self.running_enums
				msg.total_finished = self.total_counter
				msg.speed = str(cd / td)

				await self.progress_queue.put(msg)

	async def stop_agents(self):
		logger.debug('mgr stop')

		info = self.session.query(JackDawADInfo).get(self.ad_id)
		info.ldap_enumeration_state = 'FINISHED'
		self.session.commit()
		self.session.close()
		for _ in range(self.agent_cnt):
			await self.agent_in_q.put(None)

		await asyncio.sleep(1)
		for agent in self.agents:
			agent.cancel()

		if self.progress_queue is not None:
			msg = LDAPEnumeratorProgress()
			msg.type = 'LDAP'
			msg.msg_type = 'FINISHED'
			msg.adid = self.ad_id
			msg.domain_name = self.domain_name
			await self.progress_queue.put(msg)

		logger.debug('All agents finished!')

	async def run(self):
		logger.info('[+] Starting LDAP information acqusition. This might take a while...')
		
		await self.setup()
		

		if self.progress_queue is not None:
			msg = LDAPEnumeratorProgress()
			msg.type = 'LDAP'
			msg.msg_type = 'STARTED'
			msg.adid = self.ad_id
			msg.domain_name = self.domain_name
			await self.progress_queue.put(msg)

		await self.check_jobs(None)

		while True:
			try:
				res = await self.agent_out_q.get()
				await self.update_progress()
				res_type, res = res

				if res_type == LDAPAgentCommand.DOMAININFO:
					self.domaininfo_ctr += 1
					await self.store_domain(res)
				
				elif res_type == LDAPAgentCommand.USER:
					self.user_ctr += 1
					await self.store_user(res)

				elif res_type == LDAPAgentCommand.MACHINE:
					self.machine_ctr += 1
					await self.store_machine(res)

				elif res_type == LDAPAgentCommand.GROUP:
					self.group_ctr += 1
					await self.store_group(res)

				elif res_type == LDAPAgentCommand.OU:
					self.ou_ctr += 1
					await self.store_ous(res)

				elif res_type == LDAPAgentCommand.GPO:
					self.gpo_ctr += 1
					await self.store_gpo(res)

				elif res_type == LDAPAgentCommand.SPNSERVICE:
					self.spn_ctr += 1
					await self.store_spn(res)
					
				elif res_type == LDAPAgentCommand.SD:
					self.sd_ctr += 1
					await self.store_sd(res)
					
				elif res_type == LDAPAgentCommand.MEMBERSHIP:
					self.member_ctr += 1
					await self.store_membership(res)

				elif res_type == LDAPAgentCommand.TRUSTS:
					self.trust_ctr += 1
					await self.store_trust(res)

				elif res_type == LDAPAgentCommand.EXCEPTION:
					logger.warning(str(res))
					
				elif res_type.name.endswith('FINISHED'):
					t = await self.check_jobs(res_type)
					if t is True:
						break
			except Exception as e:
				logger.exception('ldap enumerator main!')
				await self.stop_agents()
				return None
			
		await self.stop_agents()
		logger.info('[+] LDAP information acqusition finished!')
		return self.ad_id
			

if __name__ == '__main__':
	from msldap.commons.url import MSLDAPURLDecoder

	import sys
	sql = sys.argv[1]
	ldap_conn_url = sys.argv[2]

	print(sql)
	print(ldap_conn_url)
	logger.setLevel(2)

	ldap_mgr = MSLDAPURLDecoder(ldap_conn_url)

	mgr = LDAPEnumeratorManager(sql, ldap_mgr)
	mgr.run()





class GraphConstruct:
	def __init__(self, ad_id, diff_name = None):
		self.ad_id = ad_id
		self.diff_name = diff_name # should be a string 'new' or 'old'
		
		self.include_nodes = []
		self.include_edges = []

		self.blacklist_sids = {'S-1-5-32-545': ''}
		self.ignoresids = {"S-1-3-0": '', "S-1-5-18": ''}


	def is_blacklisted_sid(self, sid):
		if sid in self.blacklist_sids:
			return True
		if sid[:len('S-1-5-21')] == 'S-1-5-21':
			if sid[-3:] == '513':
				return True
			
		return False

	@staticmethod
	def from_dict(d):
		g = GraphConstruct(d['ad_id'])
		g.diff_name = d['diff_name']
		
		#g.include_nodes = []
		#g.include_edges = []

		g.blacklist_sids = d['blacklist_sids']
		g.ignoresids = d['ignoresids']

		return g

	def to_dict(self):
		return {
			'ad_id' : self.ad_id,
			'diff_name' : self.diff_name, 
			'include_nodes': None, #TODO! FIX!
			'include_edges' : None, #TODO! FIX!
			'blacklist_sids' : self.blacklist_sids,
			'ignoresids' : self.ignoresids
		}


import asyncio
import logging
import json
import traceback
import ipaddress
import multiprocessing
import threading
import enum
import copy

from tqdm import tqdm
from dns import resolver, reversename
from aiosmb.commons.interfaces.share import SMBShare
from aiosmb.commons.interfaces.file import SMBFile
from aiosmb.commons.interfaces.directory import SMBDirectory
from aiosmb.commons.connection.url import SMBConnectionURL

import aiosmb
from aiosmb.commons.utils.extb import format_exc

from jackdaw.common.apq import AsyncProcessQueue
from jackdaw.dbmodel.netshare import NetShare
from jackdaw.dbmodel.netdir import NetDir
from jackdaw.dbmodel.netfile import NetFile
from jackdaw.dbmodel.netdacl import NetDACL
from jackdaw import logger
from jackdaw.dbmodel import create_db, get_session

from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.adcomp import JackDawADMachine

class SMBShareGathererCmd(enum.Enum):
	START = 'START'
	END = 'END'
	TERMINATE = 'TERMINATE'

# only using share settings, making it more granular (per dir, per file) would cause massive overload
# on some systems with a lot of files available
class SMBShareGathererSettings:
	def __init__(self, ad_id, smb_mgr, share_id, target, smbshare):
		self.ad_id = ad_id
		self.smb_mgr = smb_mgr
		self.share_id = share_id
		self.target = target
		self.share = smbshare
		
		#self.dir_only = True
		self.dir_depth = 5
		self.dir_name_filter = None
		self.dir_with_sd = True

		#self.file_with_data = False
		self.file_with_sd = True
		#self.file_name_filter = None

class SMBEnumError:
	def __init__(self, settings, error):
		self.error = error
		self.settings = settings

class SMBShareGathererResult:
	def __init__(self, result, settings):
		self.result = result
		self.settings = settings
		self.type = self.get_type()

	def get_type(self):
		if isinstance(self.result, SMBDirectory):
			return 'dir'
		else:
			return 'unknown'


class ShareGathererProcess(multiprocessing.Process):
	def __init__(self, in_q, out_q, worker_cnt):
		multiprocessing.Process.__init__(self)
		self.workers = []
		self.worker_cnt = worker_cnt
		self.in_q = in_q
		self.out_q = out_q
		self.semaphore = None

		self.main_task = None
		self.stop_task = None
	############# Enumerator worker

	async def enum_directory(self, connection, settings, directory, depth):
		try:
			await directory.list(connection)
			for name in directory.subdirs:
				if settings.dir_with_sd == True:
					await directory.subdirs[name].get_security_descriptor(connection)

			for name in directory.files:
				if settings.file_with_sd == True:
					await directory.files[name].get_security_descriptor(connection)

			res = SMBShareGathererResult(directory, settings)
			await self.out_q.coro_put(res)
			if depth == 0:
				return
			for name in directory.subdirs:
				await self.enum_directory(connection, settings, directory.subdirs[name], depth - 1)

		except asyncio.CancelledError:
			return

		except:
			traceback.print_exc()
			await self.out_q.coro_put(SMBEnumError(settings, traceback.format_exc()))

	async def enum_share(self, settings):
		try:
			connection = settings.smb_mgr.create_connection_newtarget(settings.target)
			await connection.login()
			#print('login ok!')
			await settings.share.connect(connection)
			#print('share ok!')
			await self.enum_directory(connection, settings, settings.share.subdirs[''], settings.dir_depth)
			#print('list ok!')

		except asyncio.CancelledError:
			return

		except:
			await self.out_q.coro_put(SMBEnumError(settings, traceback.format_exc()))

		finally:
			self.semaphore.release()


	################### Manager logic

	async def setup(self):
		self.semaphore = asyncio.Semaphore(self.worker_cnt)

	async def terminate(self):
		#print('terminating')
		for worker_task in self.workers:
			worker_task.cancel()

		if self.stop_task:
			self.stop_task.cancel()
		self.main_task.cancel()

	async def stop(self):
		for worker_task in self.workers:
			await worker_task
		await self.terminate()
		return

	async def main(self):
		await self.setup()
		while True:
			data_in = await self.in_q.coro_get()
			#print('process got data!')
			if data_in[0] == SMBShareGathererCmd.TERMINATE:
				await self.terminate()
				return
			elif data_in[0] == SMBShareGathererCmd.START:
				await self.semaphore.acquire()
				self.workers.append(asyncio.create_task(self.enum_share(data_in[1])))
			elif data_in[0] == SMBShareGathererCmd.END:
				self.stop_task = asyncio.create_task(self.stop())

	async def a_run(self):
		try:
			self.main_task = asyncio.create_task(self.main())
			await self.main_task
		except asyncio.CancelledError:
			return
		except:
			await self.out_q.coro_put(SMBEnumError(None, traceback.format_exc()))
		finally:
			await self.out_q.coro_put(None)
		
	def run(self):
		asyncio.run(self.a_run())

class ShareGathererManager:
	def __init__(self, settings_base, db_conn = None, db_session = None, worker_cnt = 30):
		self.settings_base = settings_base
		self.in_q = AsyncProcessQueue()
		self.out_q = AsyncProcessQueue()
		self.worker_cnt = worker_cnt
		self.targets_thread = None
		self.db_conn = db_conn
		self.db_session = db_session
		self.dir_lookup = {}
		self.exclude_shares = ['IPC$']

		self.use_progress_bar = True
		self.prg_dirs = None
		self.prg_files = None
		self.prg_errors = None
	
	def generate_targets(self):
		session = get_session(self.db_conn)
		qry = session.query(
			JackDawADMachine.sAMAccountName, NetShare.id, NetShare.netname
			).filter(JackDawADMachine.ad_id == self.settings_base.ad_id
			).filter(JackDawADMachine.id == NetShare.machine_id)

		for mname, shareid, sharename in qry.all():
			if sharename in self.exclude_shares:
				continue
			target = mname if mname[-1] != '$' else mname[:-1]
			fullpath = '\\\\%s\\%s' % (target, sharename)
			smbshare = SMBShare(fullpath = fullpath)

			#print(target)
			#print(fullpath)
			#print(smbshare)

			settings = copy.deepcopy(self.settings_base)
			settings.share_id = shareid
			settings.target = target
			settings.share = smbshare
			self.in_q.put(( SMBShareGathererCmd.START, settings))

		session.close()
		self.in_q.put(( SMBShareGathererCmd.END, None))

	def init_dbsession(self):
		if self.db_session is not None:
			return
		self.db_session = get_session(self.db_conn)
	
	def setup(self):
		self.init_dbsession()
		self.targets_thread = threading.Thread(target = self.generate_targets)
		self.targets_thread.daemon = True
		self.targets_thread.start()

		self.gatherer = ShareGathererProcess(self.in_q, self.out_q, self.worker_cnt)
		self.gatherer.start()

		if self.use_progress_bar is True:
			self.prg_dirs = tqdm(desc='Dirs', ascii = True)
			self.prg_files = tqdm(desc='files', ascii = True)
			self.prg_errors = tqdm(desc='Errors', ascii = True)

	def get_dir_id(self, unc_path, name):
		#print(name)
		if name == '' or name is None:
			return None
		# this might get extended later!
		return self.dir_lookup[unc_path]

	def store_sd(self, sd, object_type, object_id):
		if sd is None:
			return
		order_ctr = 0
		for ace in sd.Dacl.aces:
			acl = NetDACL()
			acl.object_id = object_id
			acl.object_type = object_type
			acl.owner_sid = str(sd.Owner)
			acl.group_sid = str(sd.Group)
			acl.ace_order = order_ctr
			
			order_ctr += 1
			acl.sd_control = sd.Control
			
			acl.ace_type = ace.Header.AceType.name
			acl.ace_mask = ace.Mask
			t = getattr(ace,'ObjectType', None)
			if t:
				acl.ace_objecttype = str(t)
			
			t = getattr(ace,'InheritedObjectType', None)
			if t:
				acl.ace_inheritedobjecttype = str(t)
				
			true_attr, false_attr = NetDACL.mask2attr(ace.Mask)
			
			for attr in true_attr:	
				setattr(acl, attr, True)
			for attr in false_attr:	
				setattr(acl, attr, False)
				
			true_attr, false_attr = NetDACL.hdrflag2attr(ace.Header.AceFlags)
			
			for attr in true_attr:	
				setattr(acl, attr, True)
			for attr in false_attr:	
				setattr(acl, attr, False)
			
			acl.ace_sid = str(ace.Sid)
			self.db_session.add(acl)
		
		self.db_session.commit()

	def run(self):
		self.setup()
		try:
			self.collect_results()
		except:
			print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
			traceback.print_exc()

	def collect_results(self):
		while True:
			result = self.out_q.get()
			if result is None:
				#print('enumerator finished!')
				break

			if isinstance(result, SMBEnumError):
				#something went error
				tid = result.settings
				if result.settings is not None:
					tid = result.settings.target
				logger.debug('[AIOSMBShareGatherer][Error][%s] %s' % (tid, result.error))
				if self.use_progress_bar is True:
					self.prg_errors.update()

			elif isinstance(result, SMBShareGathererResult):
				if result.type == 'dir':
					if self.use_progress_bar is True:
						self.prg_dirs.update()

					
					nd = NetDir()
					nd.share_id = result.settings.share_id
					nd.parent_id = self.get_dir_id(result.result.unc_path, result.result.name)
					nd.creation_time = result.result.creation_time
					nd.last_access_time = result.result.creation_time
					nd.last_write_time = result.result.last_write_time
					nd.change_time = result.result.change_time
					nd.unc = result.result.unc_path
					nd.name = result.result.name
					self.db_session.add(nd)
					self.db_session.commit()
					self.db_session.refresh(nd)
					self.dir_lookup[result.result.unc_path] = nd.id
					for subdir in result.result.subdirs:
						sd = result.result.subdirs[subdir]
						nsd = NetDir()
						nsd.share_id = result.settings.share_id
						nsd.parent_id = nd.id
						nsd.creation_time = sd.creation_time
						nsd.last_access_time = sd.creation_time
						nsd.last_write_time = sd.last_write_time
						nsd.change_time = sd.change_time
						nsd.unc = sd.unc_path
						nsd.name = sd.name
						self.db_session.add(nsd)
						self.db_session.commit()
						self.db_session.refresh(nsd)
						self.dir_lookup[sd.unc_path] = nd.id
						self.store_sd(sd.sid, 'dir', nd.id)

					for filename in result.result.files:
						if self.use_progress_bar is True:
							self.prg_files.update()
						f = result.result.files[filename]
						nf = NetFile()
						nf.folder_id = nd.id
						nf.creation_time = f.creation_time
						nf.last_access_time = f.last_access_time
						nf.last_write_time = f.last_write_time
						nf.change_time = f.change_time
						nf.unc = f.unc_path
						nf.size = f.size
						name = f.name
						x = f.name.rsplit('.',1)
						if len(x) > 1:
							name = x[0]
							ext = x[1]
						nf.name = name
						nf.ext = ext
						self.db_session.add(nf)
						self.db_session.commit()
						self.db_session.refresh(nf)
						self.store_sd(f.sid, 'file', nf.id)

				#elif result.type == 'file':
					

			#if result is None and error is None:
			#	logger.debug('Finished: %s' % target.ip)
			#	if self.use_progress_bar is True:
			#		self.prg_hosts.update()


if __name__ == '__main__':
	db_conn = 'sqlite:///a.db'
	create_db(db_conn)
	smb_mgr = SMBConnectionURL('smb+ntlm-password://TEST\\victim:Passw0rd!1@10.10.10.2')	
	
	ad_id = 0
	settings = SMBShareGathererSettings(ad_id, smb_mgr, None, None, None)
	ep = ShareGathererManager(settings, db_conn = db_conn, worker_cnt = 30)
	ep.run()
	print('done!')


from . import Basemodel
import datetime
from sqlalchemy import Column, Integer, String, DateTime

class NetFile(Basemodel):
	__tablename__ = 'netfile'
	
	id = Column(Integer, primary_key=True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	folder_id = Column(Integer, index=True)
	creation_time = Column(DateTime)
	last_access_time = Column(DateTime)
	last_write_time = Column(DateTime)
	change_time = Column(DateTime)
	unc = Column(String, index=True)
	size = Column(Integer, index=True)
	name = Column(String, index=True)
	ext = Column(String, index=True)

	def to_dict(self):
		return {
			'id' : self.id , 
			'fetched_at' : self.fetched_at , 
			'share_id' : self.share_id , 
			'parent_id' : self.parent_id , 
			'creation_time' : self.creation_time , 
			'last_access_time' : self.last_access_time , 
			'last_write_time' : self.last_write_time , 
			'change_time' : self.change_time , 
			'unc' : self.unc ,
			'size' : self.size ,
			'name' : self.name ,
			'ext' : self.ext ,

		}



from jackdaw.nest.graph.domain import DomainGraph

class DomainDiff:
	def __init__(self, db_conn = None, dbsession = None):
		self.db_conn = db_conn
		self.dbsession = dbsession
		self.old_graph = None
		self.new_graph = None

		self.users_added = {}
		self.users_removed = {}

		self.machines_added = {}
		self.machines_removed = {}

		self.groups_added = {}
		self.groups_removed = {}

		self.ous_added = {}
		self.ous_removed = {}

	def diff_nodes(self):
		for sid, attrs in self.new_graph.graph.nodes(data=True):
			print(sid)
			if not self.old_graph.graph.has_node(sid):
				print(attrs)
				if attrs.get('node_type') == 'user':
					self.users_added[sid] = {
						'name' : attrs['name'],
						'adid' : attrs['construct'].ad_id,
					}
				elif attrs.get('node_type') == 'machine':
					self.machines_added[sid] = {
						'name' : attrs['name'],
						'adid' : attrs['construct'].ad_id,
					}
				elif attrs.get('node_type') == 'group':
					self.groups_added[sid] = {
						'name' : attrs['name'],
						'adid' : attrs['construct'].ad_id,
					}
	
		for sid, attrs in self.old_graph.graph.nodes(data=True):
			if not self.new_graph.graph.has_node(sid):
				if attrs.get('node_type') == 'user':
					self.users_removed[sid] = {
						'name' : attrs['name'],
						'adid' : attrs['construct'].ad_id,
					}
				elif attrs.get('node_type') == 'machine':
					self.machines_removed[sid] = {
						'name' : attrs['name'],
						'adid' : attrs['construct'].ad_id,
					}
				elif attrs.get('node_type') == 'group':
					self.groups_removed[sid] = {
						'name' : attrs['name'],
						'adid' : attrs['construct'].ad_id,
					}

		node_diff = {
			'users_added' : self.users_added,
			'users_removed' : self.users_removed,
			'machines_added' : self.machines_added,
			'machines_removed' : self.machines_removed,
			'groups_added' : self.groups_added,
			'groups_removed' : self.groups_removed,
			'ous_added' : self.ous_added,
			'ous_removed' : self.ous_removed,
		}

		return node_diff

	def diff_edges(self):
		pass

	def diff_path(self, srcsid = None, dstsid = None):
		if srcsid is None and dstsid is None:
			raise Exception('Either src or dst must be provided, or both!')
		if srcsid is not None:
			if not self.old_graph.graph.has_node(srcsid):
				raise Exception('srcsid not found in old version!')
			if not self.new_graph.graph.has_node(srcsid):
				raise Exception('srcsid not found in new version!')
		if dstsid is not None:
			if not self.old_graph.graph.has_node(dstsid):
				raise Exception('dstsid not found in old version!')
			if not self.new_graph.graph.has_node(dstsid):
				raise Exception('dstsid not found in new version!')

		path_old = self.old_graph.all_shortest_paths(srcsid, dstsid)
		path_new = self.new_graph.all_shortest_paths(srcsid, dstsid)


	def diff_path_da(self):
		return {}

	def diff_path_distance(self, sid):
		if not self.old_graph.graph.has_node(sid):
			raise Exception('sid not found in old version!')
		if not self.new_graph.graph.has_node(sid):
			raise Exception('sid not found in new version!')
		old_dist = self.old_graph.distances_from_node(sid)
		new_dist = self.new_graph.distances_from_node(sid)

		merged = {}
		for dist in old_dist:
			if dist in new_dist:
				merged[dist] = new_dist[dist] - old_dist[dist]
			else:
				merged[dist] = 0 - old_dist[dist]
		for dist in new_dist:
			if dist not in new_dist:
				merged[dist] = new_dist[dist]

		return merged
				

	def construct(self, construct_old, construct_new):
		construct_old.diff_name = 'old'
		self.old_graph = DomainGraph(dbsession = self.dbsession)
		self.old_graph.construct(construct_old)

		construct_new.diff_name = 'new'
		self.new_graph = DomainGraph(dbsession = self.dbsession)
		self.new_graph.construct(construct_new)




from flask import current_app
from jackdaw.dbmodel.netfile import NetFile
from jackdaw.dbmodel.netshare import NetShare
from jackdaw.dbmodel.netdacl import NetDACL
from jackdaw.dbmodel.netdir import NetDir
from jackdaw.dbmodel.adcomp import JackDawADMachine


def get_file(domainid, fileid):
	db = current_app.db
	f = db.session.query(NetFile).get(fileid)
	return f.to_dict()

def get_dir(domainid, dirid):
	db = current_app.db
	f = db.session.query(NetDir).get(dirid)
	return f.to_dict()

def search_file_name(domainid, pattern, page, maxcnt):
	db = current_app.db
	pattern = pattern['pattern']
	res = {
		'res' : [],
		'page': {},
	}

	qry = db.session.query(
		NetFile.id, NetFile.name, NetFile.unc
		).filter(NetShare.machine_id == JackDawADMachine.id
		).filter(JackDawADMachine.ad_id == domainid
		).filter(NetDir.share_id == NetShare.id
		).filter(NetDir.id == NetFile.folder_id
		).filter(NetFile.name.like(pattern))

	files = []
	qry = qry.paginate(page = page, max_per_page = maxcnt)
	for fid, name, unc in qry.items:
		files.append({
			'id' : fid,
			'name' : name,
			'unc' : unc,    
		})

	page = dict(
		total=qry.total, 
		current_page=qry.page,
		per_page=qry.per_page
	)

	res['res'] = files
	res['page'] = page

	return res

def search_file_ext(domainid, pattern, page, maxcnt):
	db = current_app.db
	pattern = pattern['pattern']
	res = {
		'res' : [],
		'page': {},
	}

	qry = db.session.query(
		NetFile.id, NetFile.name, NetFile.unc
		).filter(NetShare.machine_id == JackDawADMachine.id
		).filter(JackDawADMachine.ad_id == domainid
		).filter(NetDir.share_id == NetShare.id
		).filter(NetDir.id == NetFile.folder_id
		).filter(NetFile.ext.like(pattern))

	files = []
	qry = qry.paginate(page = page, max_per_page = maxcnt)
	for fid, name, unc in qry.items:
		files.append({
			'id' : fid,
			'name' : name,
			'unc' : unc,    
		})

	page = dict(
		total=qry.total, 
		current_page=qry.page,
		per_page=qry.per_page
	)

	res['res'] = files
	res['page'] = page

	return res

def search_file_full(domainid, filter, page, maxcnt):
	db = current_app.db
	pattern = filter['pattern']
	ownersid = filter.get('ownersid')
	machineid = filter.get('machineid')
	size_greather = filter.get('size_greather')
	size_smaller = filter.get('size_smaller')
	created_after = filter.get('created_after')
	created_before = filter.get('created_before')
	changed_after = filter.get('changed_after')
	changed_before = filter.get('changed_before')
	ext_pattern = filter.get('extension_pattern')
	
	res = {
		'res' : [],
		'page': {},
	}

	qry = db.session.query(
		NetFile.id, NetFile.name, NetFile.unc
		).filter(NetShare.machine_id == JackDawADMachine.id
		).filter(JackDawADMachine.ad_id == domainid
		).filter(NetDir.share_id == NetShare.id
		).filter(NetDir.id == NetFile.folder_id
		).filter(NetFile.unc.like(pattern))

	if ownersid is not None:
		qry = qry.filter(NetDACL.object_id == NetFile.id
			).filter(NetDACL.owner_sid == ownersid
			)

	if machineid is not None:
		qry = qry.filter(JackDawADMachine.id == machineid)

	if created_after is not None:
		qry = qry.filter(NetFile.creation_time >= created_after)
	
	if created_after is not None:
		qry = qry.filter(NetFile.creation_time <= created_before)
	
	if changed_after is not None:
		qry = qry.filter(NetFile.change_time >= changed_after)
	
	if changed_before is not None:
		qry = qry.filter(NetFile.change_time <= changed_before)
	
	if size_greather is not None:
		qry = qry.filter(NetFile.size >= size_greather)
	
	if size_smaller is not None:
		qry = qry.filter(NetFile.size <= size_smaller)
	
	if ext_pattern is not None:
		qry = qry.filter(NetFile.ext.like(ext_pattern))


	files = []
	qry = qry.paginate(page = page, max_per_page = maxcnt)
	for fid, name, unc in qry.items:
		files.append({
			'id' : fid,
			'name' : name,
			'unc' : unc,    
		})

	page = dict(
		total=qry.total, 
		current_page=qry.page,
		per_page=qry.per_page
	)

	res['res'] = files
	res['page'] = page

	return res

def search_file_owner(domainid, ownersid, pattern, page, maxcnt):
	db = current_app.db
	pattern = pattern['pattern']
	res = {
		'res' : [],
		'page': {},
	}

	qry = db.session.query(
		NetFile.id, NetFile.name, NetFile.unc
		).filter(NetShare.machine_id == JackDawADMachine.id
		).filter(JackDawADMachine.ad_id == domainid
		).filter(NetDir.share_id == NetShare.id
		).filter(NetDir.id == NetFile.folder_id
		).filter(NetDACL.object_id == NetFile.id
		).filter(NetDACL.owner_sid == ownersid
		).filter(NetFile.unc.like(pattern))

	files = []
	qry = qry.paginate(page = page, max_per_page = maxcnt)
	for fid, name, unc in qry.items:
		files.append({
			'id' : fid,
			'name' : name,
			'unc' : unc,    
		})

	page = dict(
		total=qry.total, 
		current_page=qry.page,
		per_page=qry.per_page
	)

	res['res'] = files
	res['page'] = page

	return res


import zipfile
import json
import codecs
import pprint

from jackdaw.dbmodel import *
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.adcomp import JackDawADMachine
from jackdaw.dbmodel.aduser import JackDawADUser
from jackdaw.dbmodel.adgroup import JackDawADGroup
from jackdaw.dbmodel.adou import JackDawADOU

class BHImport:
	def __init__(self, db_conn = None, db_session = None):
		self.zipfile = None
		self.files = None
		self.db_conn = db_conn
		self.db_session = db_session

		self.is_zip = False
		self.fd = {}
		self.ads = {}
		self.adn = {} #name -> ad_id

		#self.setup_db()

	def setup_db(self):
		if self.db_session is None:
			self.db_session = get_session(self.db_conn)

	def import_machines(self):
		print('Importing machines!')
		for machine in self.get_file('computers')['computers']:
			#pprint.pprint(machine)
			#input()
			
			m = JackDawADMachine()
			m.ad_id = self.ads[machine['Properties']['objectsid'].rsplit('-',1)[0]]
			m.sAMAccountName = machine['Name'].split('.', 1)[0] + '$'
			m.objectSid = machine['Properties']['objectsid']
			m.description = machine['Properties']['description']
			m.operatingSystemVersion  = machine['Properties']['operatingsystem']

			self.db_session.add(m)
		self.db_session.commit()

	def import_users(self):
		print('Importing users!')
		for user in self.get_file('users')['users']:
			#pprint.pprint(user)
			#input()
			
			m = JackDawADUser()
			m.ad_id = self.ads[user['Properties']['objectsid'].rsplit('-',1)[0]]
			m.name = user['Name'].split('@', 1)[0]
			m.objectSid = user['Properties']['objectsid']
			m.description = user['Properties']['description']
			m.displayName  = user['Properties']['displayname']
			m.email  = user['Properties']['email']

			self.db_session.add(m)
		self.db_session.commit()

	def import_sessions(self):
		print('Importing sessions!')
		for session in self.get_file('sessions')['sessions']:
			#pprint.pprint(session)
			#input()
			try:
				if session['ComputerName'].startswith('['):
					continue
				ad_name = session['UserName'].rsplit('@', 1)[1]
				cname = session['ComputerName'] + '$'
				if session['ComputerName'].find('.') != -1:
					cname = session['ComputerName'].split('.', 1)[0] + '$'

				qry = self.db_session.query(
					JackDawADMachine.id
					).filter_by(ad_id = self.adn[ad_name]
					).filter(JackDawADMachine.sAMAccountName == cname
					)
				machine_id = qry.first()
				if machine_id is None:
					raise Exception('Could not find machine!')
				m = NetSession()
				m.machine_id = machine_id[0]
				m.username = session['UserName'].split('@', 1)[0]

				self.db_session.add(m)
			except Exception as e:
				#print(e)
				#pprint.pprint(session)
				#input()
				continue
		self.db_session.commit()

	def import_ous(self):
		print('Importing ous!')
		for ou in self.get_file('ous')['ous']:
			#pprint.pprint(groups)
			#input()
			try:
				ad_name = ou['Name'].rsplit('@', 1)[1]
				m = JackDawADOU()
				m.ad_id = self.adn[ad_name]
				m.name = ou['Name'].split('@', 1)[0]
				m.objectSid = ou['Properties']['objectsid']
				m.description = ou['Properties'].get('description', None)

				self.db_session.add(m)
			except Exception as e:
				print(e)
				pprint.pprint(ou)
				input()
				continue
		self.db_session.commit()

	def import_domains(self):
		print('Importing domains!')
		for domain in self.get_file('domains')['domains']:#['computers']:
			#pprint.pprint(domain)
			#input()

			di = JackDawADInfo()
			di.name = domain['Name']
			di.objectSid = domain['Properties']['objectsid']

			self.db_session.add(di)
			self.db_session.commit()
			self.db_session.refresh(di)
			self.ad_id = di.id

			self.ads[di.objectSid] = di.id
			self.adn[di.name] = di.id

	def import_gpos(self):
		print('Importing gpos!')
		for gpo in self.get_file('gpos')['gpos']:
			pprint.pprint(gpo)
			input()
			try:
				ad_name = ou['Name'].rsplit('@', 1)[1]
				m = JackDawADOU()
				m.ad_id = self.adn[ad_name]
				m.name = ou['Name'].split('@', 1)[0]
				m.objectSid = ou['Properties']['objectsid']
				m.description = ou['Properties'].get('description', None)

				self.db_session.add(m)
			except Exception as e:
				print(e)
				pprint.pprint(ou)
				input()
				continue
		self.db_session.commit()

	def import_groups(self):
		print('Importing groups!')
		for groups in self.get_file('groups')['groups']:
			#pprint.pprint(groups)
			#input()
			try:
				ad_name = groups['Name'].rsplit('@', 1)[1]
				m = JackDawADGroup()
				m.ad_id = self.adn[ad_name]
				m.name = groups['Name'].split('@', 1)[0]
				m.objectSid = groups['Properties']['objectsid']
				m.description = groups['Properties'].get('description', None)

				self.db_session.add(m)
			except Exception as e:
				print(e)
				pprint.pprint(groups)
				input()
				continue
		self.db_session.commit()

	def get_file(self, filetype):
		if self.is_zip is True:
			with zipfile.ZipFile(filepath, 'r') as zf:
				with zf.open(self.fd[filetype]) as data:
					return json.load(data)

	@staticmethod
	def from_zipfile(filepath):
		bh = BHImport()
		if not zipfile.is_zipfile(filepath):
			raise Exception('The file on this path doesnt look like a valid zip file! %s' % filepath)
		
		bh.is_zip = True
		zip = zipfile.ZipFile(filepath, 'r')
		for filename in zip.namelist():
			if filename.find('_computers.json') != -1:
				bh.fd['computers'] = filename
			elif filename.find('_domains.json') != -1:
				bh.fd['domains'] = filename
			elif filename.find('_gpos.json') != -1:
				bh.fd['gpos'] = filename
			elif filename.find('_groups.json') != -1:
				bh.fd['groups'] = filename
			elif filename.find('_ous.json') != -1:
				bh.fd['ous'] = filename
			elif filename.find('_sessions.json') != -1:
				bh.fd['sessions'] = filename
			elif filename.find('_users.json') != -1:
				bh.fd['users'] = filename

		return bh

	def from_folder(self, folderpath):
		pass

	def run(self):
		#DO NOT CHANGE THIS ORDER!!!!
		self.setup_db()
		self.import_domains()
		#self.import_groups()
		#self.import_machines()
		#self.import_users()
		#self.import_sessions()
		self.import_gpos()


		#self.import_ous() #not working!


if __name__ == '__main__':
	import sys
	db_conn = 'sqlite:///bhtest.db'
	filepath = sys.argv[1]

	create_db(db_conn)
	
	bh = BHImport.from_zipfile(filepath)
	bh.db_conn = db_conn
	bh.run()




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import multiprocessing as mp
import threading
import enum
import gzip
import json

from sqlalchemy import func
from sqlalchemy import not_, and_, or_, case
from sqlalchemy.orm import load_only
import networkx as nx
from networkx.readwrite import json_graph

from jackdaw.dbmodel import get_session
from jackdaw.dbmodel.spnservice import JackDawSPNService
from jackdaw.dbmodel.addacl import JackDawADDACL
from jackdaw.dbmodel.adsd import JackDawSD
from jackdaw.dbmodel.adgroup import JackDawADGroup
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.aduser import JackDawADUser
from jackdaw.dbmodel.adcomp import JackDawADMachine
from jackdaw.dbmodel.adou import JackDawADOU
from jackdaw.dbmodel.usergroup import JackDawGroupUser
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.tokengroup import JackDawTokenGroup
from jackdaw.dbmodel.adgpo import JackDawADGPO
from jackdaw.dbmodel.constrained import JackDawMachineConstrainedDelegation, JackDawUserConstrainedDelegation
from jackdaw.dbmodel.adgplink import JackDawADGplink
from jackdaw.dbmodel.netsession import NetSession
from jackdaw.dbmodel.localgroup import LocalGroup
from jackdaw.dbmodel.credential import Credential
from jackdaw.wintypes.well_known_sids import get_name_or_sid, get_sid_for_name
from jackdaw.wintypes.lookup_tables import *
from jackdaw.nest.graph.graphdata import *
from jackdaw import logger
from jackdaw.utils.encoder import UniversalEncoder
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR
import base64
from tqdm import tqdm

from jackdaw.nest.graph.construct import GraphConstruct
class GraphDecoder(json.JSONDecoder):
	def __init__(self, *args, **kwargs):
		json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
	def object_hook(self, dct):
		if 'construct' in dct:
			dct['construct'] = GraphConstruct.from_dict(dct['construct'])
		return dct

def windowed_query(q, column, windowsize, is_single_entity = True):
	""""Break a Query into chunks on a given column."""

	#single_entity = q.is_single_entity
	q = q.add_column(column).order_by(column)
	last_id = None

	while True:
		subq = q
		if last_id is not None:
			subq = subq.filter(column > last_id)
		chunk = subq.limit(windowsize).all()
		if not chunk:
			break
		last_id = chunk[-1][-1]
		for row in chunk:
			if is_single_entity is True:
				yield row[0]
			else:
				yield row[0:-1]


def ace_applies(ace_guid, object_class):
	'''
	Checks if an ACE applies to this object (based on object classes).
	Note that this function assumes you already verified that InheritedObjectType is set (via the flag).
	If this is not set, the ACE applies to all object types.
	'''
	try:
		our_ace_guid = OBJECTTYPE_GUID_MAP[object_class]
	except KeyError:
		return False
	if ace_guid == our_ace_guid:
		return True
	# If none of these match, the ACE does not apply to this object
	return False

class NodeType(enum.Enum):
	USER = 'USER'
	GROUP = 'GROUP'
	MACHINE = 'MACHINE'
	
class EdgeType(enum.Enum):
	MEMBER_GROUP = 'MEMBER_GROUP'
	MEMBER_USER = 'MEMBER_USER'
	SESSION = 'SESSION'
	ACL = 'ACL'
	LOCALGROUP = 'LOCALGROUP'
	PWSHARE = 'PWSHARE'
	CUSTOM = 'CUSTOM'
	DELEGATION_CONSTRAINED = 'DELEGATION_CONSTRAINED'
	DELEGATION_UNCONSTRAINED = 'DELEGATION_UNCONSTRAINED'

def short_node_gen(graph, inqueue, dst_sid, procno):
	"""
	Adding all nodes from the graph to the inqueue. At the end it adds terminating object in the amount of proccount.
	This function will be threaded. 
	"""
	for node in graph.nodes:
		if node == dst_sid:
			continue
		inqueue.put(node)
			
	for _ in range(procno):
		inqueue.put(None)

def short_worker(inqueue, outqueue, graph, dst_sid):
	"""
	Calculates the shortest parth for a given destination node
	This function is "multiprocessed"
	"""
	while True:
		node = inqueue.get()
		if node is None:
			outqueue.put(None)
			return
		try:
			for path in nx.all_shortest_paths(graph, source = node, target = dst_sid):
				outqueue.put(path)
									
		except nx.exception.NetworkXNoPath:
			continue


def acl_calc_gen(session, adid, inqueue, procno):
	total = session.query(func.count(JackDawSD.id)).filter_by(ad_id = adid).scalar()

	q = session.query(JackDawSD).filter_by(ad_id = adid)

	for adsd in tqdm(windowed_query(q, JackDawSD.id, 1000), total=total):
		sd = SECURITY_DESCRIPTOR.from_bytes(base64.b64decode(adsd.sd))
		
		order_ctr = 0
		for ace in sd.Dacl.aces:
			acl = JackDawADDACL()
			acl.ad_id = adsd.ad_id
			acl.object_type = adsd.object_type
			acl.object_type_guid = OBJECTTYPE_GUID_MAP.get(adsd.object_type)
			acl.owner_sid = str(sd.Owner)
			acl.group_sid = str(sd.Group)
			acl.ace_order = order_ctr
			
			order_ctr += 1
			acl.guid = str(adsd.guid)
			if adsd.sid:
				acl.sid = str(adsd.sid)
			#if sd.cn:
			#	acl.cn = sd.cn
			#if sd.distinguishedName:
			#	acl.dn = str(sd.distinguishedName)
			acl.sd_control = sd.Control
			
			acl.ace_type = ace.AceType.name
			acl.ace_mask = ace.Mask
			t = getattr(ace,'ObjectType', None)
			if t:
				acl.ace_objecttype = str(t)
			
			t = getattr(ace,'InheritedObjectType', None)
			if t:
				acl.ace_inheritedobjecttype = str(t)
				
			true_attr, false_attr = JackDawADDACL.mask2attr(ace.Mask)
			
			for attr in true_attr:	
				setattr(acl, attr, True)
			for attr in false_attr:	
				setattr(acl, attr, False)
				
			true_attr, false_attr = JackDawADDACL.hdrflag2attr(ace.AceFlags)
			
			for attr in true_attr:	
				setattr(acl, attr, True)
			for attr in false_attr:	
				setattr(acl, attr, False)
			
			acl.ace_sid = str(ace.Sid)
		
			inqueue.put(acl)
	#adinfo = session.query(JackDawADInfo).get(adid)
	#for acl in adinfo.objectacls:
	#	inqueue.put(acl)

	for _ in range(procno):
		inqueue.put(None)

def acl_calc_mp(inqueue, outqueue, construct):
	while True:
		acl = inqueue.get()
		if acl is None:
			outqueue.put(None)
			return
		
		if acl.owner_sid not in construct.ignoresids:
			outqueue.put((acl.owner_sid, acl.sid, 'Owner'))
				
		if acl.ace_sid in construct.ignoresids:
			continue
			
		if acl.ace_type not in ['ACCESS_ALLOWED_ACE_TYPE','ACCESS_ALLOWED_OBJECT_ACE_TYPE']:
			continue

		if acl.ace_type == 'ACCESS_ALLOWED_ACE_TYPE':
			if acl.ace_mask_generic_all == True:
				outqueue.put((acl.ace_sid, acl.sid, 'GenericALL'))
			
			if acl.ace_mask_generic_write == True:
				outqueue.put((acl.ace_sid, acl.sid, 'GenericWrite'))
				
			if acl.ace_mask_write_owner == True:
				outqueue.put((acl.ace_sid, acl.sid, 'WriteOwner'))
				
			if acl.ace_mask_write_dacl == True:
				outqueue.put((acl.ace_sid, acl.sid, 'WriteDacl'))
				
			if acl.object_type in ['user', 'domain'] and acl.ace_mask_control_access == True:
				outqueue.put((acl.ace_sid, acl.sid, 'ExtendedRightALL'))
		
		if acl.ace_type == 'ACCESS_ALLOWED_OBJECT_ACE_TYPE':
			if acl.ace_hdr_flag_inherited == True and acl.ace_hdr_flag_inherit_only == True:
				continue
			
			if acl.ace_hdr_flag_inherited == True and acl.ace_inheritedobjecttype is not None:
				if not ace_applies(acl.ace_inheritedobjecttype, acl.object_type):
					continue
			
			if any([acl.ace_mask_generic_all, acl.ace_mask_write_dacl, acl.ace_mask_write_owner, acl.ace_mask_generic_write]):
				if acl.ace_objecttype is not None and not ace_applies(acl.ace_objecttype, acl.object_type):
					continue
				
				if acl.ace_mask_generic_all == True:
					outqueue.put((acl.ace_sid, acl.sid, 'GenericALL'))
					continue
		
				if acl.ace_mask_generic_write == True:
					outqueue.put((acl.ace_sid, acl.sid, 'GenericWrite'))
					if acl.object_type != 'domain':
						continue
					
				if acl.ace_mask_write_dacl == True:
					outqueue.put((acl.ace_sid, acl.sid, 'WriteDacl'))
					
				if acl.ace_mask_write_owner == True:
					outqueue.put((acl.ace_sid, acl.sid, 'WriteOwner'))
					
			if acl.ace_mask_write_prop == True:
				if acl.object_type in ['user','group'] and acl.ace_objecttype is None:
					outqueue.put((acl.ace_sid, acl.sid, 'GenericWrite'))
					
				if acl.object_type == 'group' and acl.ace_objecttype == 'bf9679c0-0de6-11d0-a285-00aa003049e2':
					outqueue.put((acl.ace_sid, acl.sid, 'AddMember'))
					
				
		
			if acl.ace_mask_control_access == True:
				if acl.object_type in ['user','group'] and acl.ace_objecttype is None:
					outqueue.put((acl.ace_sid, acl.sid, 'ExtendedAll'))
				
				if acl.object_type == 'domain' and acl.ace_objecttype == '1131f6ad-9c07-11d1-f79f-00c04fc2dcd2':
					# 'Replicating Directory Changes All'
					outqueue.put((acl.ace_sid, acl.sid, 'GetChangesALL'))
						
				if acl.object_type == 'domain' and acl.ace_objecttype == '1131f6aa-9c07-11d1-f79f-00c04fc2dcd2':
					# 'Replicating Directory Changes'
					outqueue.put((acl.ace_sid, acl.sid, 'GetChanges'))
						
				if acl.object_type == 'user' and acl.ace_objecttype == '00299570-246d-11d0-a768-00aa006e0529':
					# 'Replicating Directory Changes'
					outqueue.put((acl.ace_sid, acl.sid, 'User-Force-Change-Password'))
		

class DomainGraph:
	def __init__(self, db_conn = None, dbsession = None):
		self.db_conn = db_conn
		self.dbsession = dbsession
		self.constructs = {}
		self.graph = nx.DiGraph()
		self.domain_sid = None

	def to_gzip(self, filename = 'test.gzip'):
		gd = json_graph.node_link_data(self.graph)
		gd['domain_sid'] = self.domain_sid

		with gzip.open(filename, 'wt', encoding="utf-8") as zipfile:
			json.dump(gd, zipfile, cls = UniversalEncoder)
	
	@staticmethod
	def from_gzip_stream(stream):
		domain_graph = DomainGraph()
		with gzip.GzipFile(fileobj=stream, mode='r') as zipfile:
			gd = json.load(zipfile, cls=GraphDecoder)
		#print(gd)
		domain_graph.domain_sid = gd['domain_sid']
		del gd['domain_sid']
		domain_graph.graph = json_graph.node_link_graph(gd)

		return domain_graph


	def get_session(self):
		if self.db_conn is not None:
			return get_session(self.db_conn)
		elif self.dbsession is not None:
			return self.dbsession
		else:
			raise Exception('Either db_conn or dbsession MUST be supplied!')

	######################################################################################################################################
	#################################################      PATH CALCULATION PART         #################################################
	######################################################################################################################################
			
	def sid2cn(self, sid, throw = False):
		session = self.get_session()
		tsid = session.query(JackDawTokenGroup.cn).filter(JackDawTokenGroup.sid == sid).first()
		#print('sid2cn: %s' % tsid)
		if not tsid:
			t = str(get_name_or_sid(str(sid)))
			if t == sid and throw == True:
				raise Exception('No CN found for SID = %s' % repr(sid))
			return t
		return tsid[0]
	
	def cn2sid(self, cn, throw = False, domain_sid = None):
		sid = get_sid_for_name(cn, domain_sid)
		
		session = self.get_session()
		tsid = session.query(JackDawTokenGroup.sid).filter(JackDawTokenGroup.cn == cn).first()
		#print(tsid)
		if not tsid:
			if throw == True:
				raise Exception('No SID found for CN = %s' % repr(cn))
			return cn
		return tsid[0]
			
	def show_all(self):
		"""
		Returns all nodes and edges from the graph
		You really don't want this on a larger graph, mostly used for testing purposes
		"""
		data = GraphData()
		for sid in self.graph.nodes:
			data.add_node(
				sid, 
				self.graph.nodes[sid].get('name', self.sid2cn(sid)), 
				self.graph.nodes[sid]['construct'].ad_id,
				self.graph.nodes[sid].get('node_type')
			)

		for edge in [e for e in self.graph.edges]:
			data.add_edge(edge[0], edge[1])

		return data
	
	def get_node(self, nodeid = None):
		
		if nodeid is None:
			nodes = []
			for gid, props in list(self.graph.nodes(data=True)):
				nodes.append(GraphNode(gid, gid, props['construct'].ad_id )) #properties = {}))
			return nodes

		if nodeid not in self.graph.nodes:
			return None
		gid, props = self.graph.nodes[nodeid]
		return GraphNode(gid, gid, props['construct'].ad_id, properties = {})
			
	def __add_path(self, network, path):
		"""
		Adds the path to the representational network
		Path is a list of sids (nodes), so we need to find the edges matching
		"""
		for d, sid in enumerate(path):
			network.add_node(
				sid, 
				name = self.graph.nodes[sid].get('name', self.sid2cn(sid)), 
				node_type = self.graph.nodes[sid].get('node_type'),
				domainid = self.graph.nodes[sid]['construct'].ad_id
			)
			network.nodes[sid].set_distance(d)
		
		for i in range(len(path) - 1):
			for edge in self.graph.edges([path[i], ], data=True):
				if edge[1] == path[i + 1]:
					name = edge[2].get('label', None)
					network.add_edge(path[i], path[i + 1], label=name)

	def distances_from_node(self, dst_sid):
		
		distances = {} #distance -> occurrence of distance
		for node in self.graph.nodes:
			try:
				for path in nx.all_shortest_paths(self.graph, source = node, target = dst_sid):
					distance = len(path)
					if distance not in distances:
						distances[distance] = 0
					distances[distance] += 1
			except nx.exception.NetworkXNoPath:
				continue

		return distances

	def all_shortest_paths(self, src_sid = None, dst_sid = None):
		nv = GraphData()
		
		if not src_sid and not dst_sid:
			raise Exception('Either source or destination MUST be specified')
		
		elif not src_sid and dst_sid:
			try:
				#for each node we calculate the shortest path to the destination node, silently skip the ones who do not have path to dst
				inqueue = mp.Queue()
				outqueue = mp.Queue()
				procno = mp.cpu_count()
				logger.debug('[DST_CALC] Starting processes')
				procs = [mp.Process(target = short_worker, args = (inqueue, outqueue, self.graph, dst_sid)) for i in range(procno)]			
				for proc in procs:
					proc.daemon = True
					proc.start()
				logger.debug('[DST_CALC] Starting generator thread')
				node_gen_th = threading.Thread(target = short_node_gen, args = (self.graph, inqueue, dst_sid, procno))
				node_gen_th.daemon = True
				node_gen_th.start()

				p_cnt = 0
				while True:
					path = outqueue.get()
					if path is None:
						procno -= 1
						logger.debug('[DST_CALC] Proc X - Finished!')
						if procno == 0:
							break
						continue
					self.__add_path(nv, path)
					p_cnt += 1

				logger.debug('[DST_CALC] Found %s paths to dst node %s' % (p_cnt, dst_sid))

				logger.debug('[DST_CALC] joining processes')
				for proc in procs:
					proc.join()
				logger.debug('[DST_CALC] Finished!')

			except:
				logger.exception('[DST_CALC]')
			
		elif src_sid and not dst_sid:
			#for each node we calculate the shortest path to the destination node, silently skip the ones who do not have path to dst
			
			for node in self.graph.nodes:
				if node == src_sid:
					continue
				try:
					for path in nx.all_shortest_paths(self.graph, source = src_sid, target = node):
						self.__add_path(nv, path)
									
				except nx.exception.NetworkXNoPath:
					continue
					
		else:
			#for each node we calculate the shortest path to the destination node, silently skip the ones who do not have path to dst
			for path in nx.all_shortest_paths(self.graph, source = src_sid, target = dst_sid):
				self.__add_path(nv, path)
		
		return nv
	

	######################################################################################################################################
	#################################################      GRAPH CALCULATION PART        #################################################
	######################################################################################################################################

	def add_sid_to_node(self, node, node_type, construct, name = None):
		if construct.is_blacklisted_sid(node):
			return
		if not name:
			name = str(get_name_or_sid(str(node)))
		
		#this presence-filter is important, as we will encounter nodes that are known and present in the graph already
		#ald later will be added via tokengroups as unknown
		if node not in self.graph.nodes:
			self.graph.add_node(str(node), name=name, node_type=node_type, construct = construct)
		
			
	def add_edge(self, sid_src, sid_dst, construct, label = None, weight = 1):
		if not sid_src or not sid_dst:
			return
		if construct.is_blacklisted_sid(sid_src) or construct.is_blacklisted_sid(sid_dst):
			return
			
		self.add_sid_to_node(sid_src, 'unknown', construct)
		self.add_sid_to_node(sid_dst, 'unknown', construct)
			
		self.graph.add_edge(sid_src, sid_dst, label = label, weight = weight)
		
		#self.network_visual.add_edge(sid_src, sid_dst, label = label, weight = weight)


	def calc_acl_edges_mp(self, session, adid, construct):
		try:
			#ACE edges calc with multiprocessing
			inqueue = mp.Queue()
			outqueue = mp.Queue()
			procno = mp.cpu_count()
			logger.debug('[ACL] Starting processes')
			procs = [mp.Process(target = acl_calc_mp, args = (inqueue, outqueue, construct)) for i in range(procno)]			
			for proc in procs:
				proc.daemon = True
				proc.start()
			logger.debug('[ACL] Starting generator thread')
			acl_gen_th = threading.Thread(target = acl_calc_gen, args = (session, adid, inqueue, procno))
			acl_gen_th.daemon = True
			acl_gen_th.start()

			p_cnt = 0
			while True:
				res = outqueue.get()
				if res is None:
					procno -= 1
					logger.debug('[ACL] Proc X - Finished!')
					if procno == 0:
						break
					continue
				ace_sid, sid, label = res
				self.add_edge(ace_sid, sid, construct, label=label)
				p_cnt += 1

			logger.debug('[ACL] Added %s edges' % (p_cnt))

			logger.debug('[ACL] joining processes')
			for proc in procs:
				proc.join()
			logger.debug('[ACL] Finished!')

		except:
			logger.exception('[ACL]')
	
	def calc_acl_edges(self, session, construct):
		logger.debug('Adding ACL edges - single thread')
		cnt = 0
		q = session.query(JackDawADDACL).filter_by(ad_id = construct.ad_id)

		for acl in windowed_query(q, JackDawADDACL.id, 1000):
			#input(acl[0])
			#acl = acl[0]
			#cnt += 1
			#if cnt % 10000 == 0:
			#	print(cnt)
			#continue
			##SUPER-IMPORTANT!!!!
			##THE ACCESS RIGHTS CALCULATIONS ARE FLAWED (JUST LIKE IN BLOODHOUND)
			##CORRECT WAY WOULD BE TO TRAVERSE ALL ACES IN A GIVEN ACL IN THE ORDER OF "ACE-ODER" AND CHECK FOR DENY POLICIES AS WELL!!!
			##TODO: CALCULATE THE EFFECTIVE PERMISSIONS IN A CORRECT MANNER :)
			
			##TODO: The DB is designed that most of these calculations can (and should) be offloaded to the DB itself
			##      Currently it's in python bc I'm laze and Dirkjan already wrote the major part
			##      https://github.com/fox-it/BloodHound.py/blob/0d3897ba4cc00818c0fc3e01fa3a91fc42e799e2/bloodhound/enumeration/acls.py
			
			if acl.owner_sid not in construct.ignoresids:
				self.add_edge(acl.owner_sid, acl.sid, construct, label='Owner')
				
			if acl.ace_sid in construct.ignoresids:
				continue
			
			if acl.ace_type in ['ACCESS_ALLOWED_ACE_TYPE','ACCESS_ALLOWED_OBJECT_ACE_TYPE']:
				if acl.ace_type == 'ACCESS_ALLOWED_ACE_TYPE':
					if acl.ace_mask_generic_all == True:
						self.add_edge(acl.ace_sid, acl.sid, construct, label='GenericALL')
						cnt += 1
					
					if acl.ace_mask_generic_write == True:
						self.add_edge(acl.ace_sid, acl.sid, construct, label='GenericWrite')
						cnt += 1
						
					if acl.ace_mask_write_owner == True:
						self.add_edge(acl.ace_sid, acl.sid, construct, label='WriteOwner')
						cnt += 1
						
					if acl.ace_mask_write_dacl == True:
						self.add_edge(acl.ace_sid, acl.sid, construct, label='WriteDacl')
						cnt += 1
						
					if acl.object_type in ['user', 'domain'] and acl.ace_mask_control_access == True:
						self.add_edge(acl.ace_sid, acl.sid, construct, label='ExtendedRightALL')
						cnt += 1
				
				if acl.ace_type == 'ACCESS_ALLOWED_OBJECT_ACE_TYPE':
					if acl.ace_hdr_flag_inherited == True and acl.ace_hdr_flag_inherit_only == True:
						continue
					
					if acl.ace_hdr_flag_inherited == True and acl.ace_inheritedobjecttype is not None:
						if not ace_applies(acl.ace_inheritedobjecttype, acl.object_type):
							continue
					
					if any([acl.ace_mask_generic_all, acl.ace_mask_write_dacl, acl.ace_mask_write_owner, acl.ace_mask_generic_write]):
						if acl.ace_objecttype is not None and not ace_applies(acl.ace_objecttype, acl.object_type):
							continue
						
						if acl.ace_mask_generic_all == True:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='GenericALL')
							cnt += 1
							continue
				
						if acl.ace_mask_generic_write == True:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='GenericWrite')
							cnt += 1
							if acl.object_type != 'domain':
								continue
							
						if acl.ace_mask_write_dacl == True:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='WriteDacl')
							cnt += 1
							
						if acl.ace_mask_write_owner == True:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='WriteOwner')
							cnt += 1
							
					if acl.ace_mask_write_prop == True:
						if acl.object_type in ['user','group'] and acl.ace_objecttype is None:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='GenericWrite')
							cnt += 1
							
						if acl.object_type == 'group' and acl.ace_objecttype == 'bf9679c0-0de6-11d0-a285-00aa003049e2':
							self.add_edge(acl.ace_sid, acl.sid, construct, label='AddMember')
							cnt += 1
							
						
				
					if acl.ace_mask_control_access == True:
						if acl.object_type in ['user','group'] and acl.ace_objecttype is None:
							self.add_edge(acl.ace_sid, acl.sid, construct, label='ExtendedAll')
							cnt += 1
						
						if acl.object_type == 'domain' and acl.ace_objecttype == '1131f6ad-9c07-11d1-f79f-00c04fc2dcd2':
							# 'Replicating Directory Changes All'
							self.add_edge(acl.ace_sid, acl.sid, construct, label='GetChangesALL')
							cnt += 1
								
						if acl.object_type == 'domain' and acl.ace_objecttype == '1131f6aa-9c07-11d1-f79f-00c04fc2dcd2':
							# 'Replicating Directory Changes'
							self.add_edge(acl.ace_sid, acl.sid, construct, label='GetChanges')
							cnt += 1
								
						if acl.object_type == 'user' and acl.ace_objecttype == '00299570-246d-11d0-a768-00aa006e0529':
							# 'Replicating Directory Changes'
							self.add_edge(acl.ace_sid, acl.sid, construct, label='User-Force-Change-Password')
							cnt += 1
		
		logger.debug('Added %s ACL edges' % cnt)

	def construct(self, construct):
		"""
		Fills the network graph from database to memory
		"""
		#self.ad_id = ad_id
		session = self.get_session()
		adinfo = session.query(JackDawADInfo).get(construct.ad_id)
		
		self.domain_sid = str(adinfo.objectSid)

		#self.calc_acl_edges(session, construct)

		#adding group nodes
		logger.debug('Adding group nodes')
		cnt = 0
		for group in adinfo.groups:
			self.add_sid_to_node(group.sid, 'group', construct, name=group.name)
			cnt += 1
		
		logger.debug('Added %s group nodes' % cnt)
		
		logger.debug('Adding user nodes')
		cnt = 0
		for user in adinfo.users:
			self.add_sid_to_node(user.objectSid, 'user', construct, name=user.sAMAccountName)
			cnt += 1
		
		logger.debug('Added %s user nodes' % cnt)
				
		logger.debug('Adding machine nodes')
		cnt = 0
		for user in adinfo.computers:
			self.add_sid_to_node(user.objectSid, 'machine', construct, name=user.sAMAccountName)
			cnt += 1
		logger.debug('Added %s machine nodes' % cnt)

		logger.debug('Adding hassession edges')
		cnt = 0
		for res in session.query(JackDawADUser.objectSid, JackDawADMachine.objectSid).filter(NetSession.username == JackDawADUser.sAMAccountName).filter(NetSession.source == JackDawADMachine.sAMAccountName).distinct(NetSession.username):
			self.add_edge(res[0], res[1], construct, label='hasSession')
			self.add_edge(res[1], res[0], construct, label='hasSession')
			cnt += 2
		logger.debug('Added %s hassession edges' % cnt)
		
		logger.debug('Adding localgroup edges')
		cnt = 0
		for res in session.query(JackDawADUser.objectSid, JackDawADMachine.objectSid, LocalGroup.groupname
					).filter(JackDawADMachine.id == LocalGroup.machine_id
					).filter(JackDawADMachine.ad_id == construct.ad_id
					).filter(JackDawADUser.ad_id == construct.ad_id
					).filter(JackDawADUser.objectSid == LocalGroup.sid
					).all():
			label = None
			if res[2] == 'Remote Desktop Users':
				label = 'canRDP'
				weight = 1
					
			elif res[2] == 'Distributed COM Users':
				label = 'executeDCOM'
				weight = 1
					
			elif res[2] == 'Administrators':
				label = 'adminTo'
				weight = 1
					
			self.add_edge(res[0], res[1], construct, label=label, weight = weight)
			cnt += 1

		logger.debug('Added %s localgroup edges' % cnt)

		# TODO: implement this!	
		#if self.show_constrained_delegations == True:
		#	pass
			
		# TODO: implement this!	
		#if self.show_unconstrained_delegations == True:
		#	pass

		# TODO: implement this!	
		#for relation in construct.custom_relations:
		#	relation.calc()
		#	self.add_edge(res.sid, res.target_sid)
			
		#print('adding membership edges')
		#adding membership edges
		logger.debug('Adding membership edges')
		cnt = 0

		q = session.query(JackDawTokenGroup).filter_by(ad_id = construct.ad_id)

		for tokengroup in windowed_query(q, JackDawTokenGroup.id, 10000):
		#for tokengroup in adinfo.group_lookups:
			self.add_sid_to_node(tokengroup.sid, 'unknown', construct)
			self.add_sid_to_node(tokengroup.member_sid, 'unknown', construct)
				
			if tokengroup.is_user == True:
				try:
					self.add_edge(tokengroup.sid, tokengroup.member_sid, construct, label='member')
					cnt += 1
				except AssertionError:
					logger.exception()
			elif tokengroup.is_machine == True:
				try:
					self.add_edge(tokengroup.sid, tokengroup.member_sid, construct, label='member')
					cnt += 1
				except AssertionError:
					logger.exception()
			elif tokengroup.is_group == True:
				try:
					self.add_edge(tokengroup.sid, tokengroup.member_sid, construct, label='member')
					cnt += 1
				except AssertionError:
					logger.exception()
		
		logger.debug('Added %s membership edges' % cnt)
		
		#adding ACL edges
		#self.calc_acl_edges(session, construct)
		#self.calc_acl_edges(adinfo, construct)
		self.calc_acl_edges_mp(session, construct.ad_id, construct)

		logger.info('Adding password sharing edges')
		cnt = 0
		def get_sid_by_nthash(ad_id, nt_hash):
			return session.query(JackDawADUser.objectSid, Credential.username
				).filter_by(ad_id = ad_id
				).filter(Credential.username == JackDawADUser.sAMAccountName
				).filter(Credential.nt_hash == nt_hash
				)

		dup_nthashes_qry = session.query(Credential.nt_hash
					).filter(Credential.history_no == 0
					).filter(Credential.ad_id == construct.ad_id
					   ).filter(Credential.username != 'NA'
					   ).filter(Credential.domain != '<LOCAL>'
					).group_by(
						Credential.nt_hash
					).having(
						func.count(Credential.nt_hash) > 1
					)

		for res in dup_nthashes_qry.all():
			sidd = {}
			for sid, _ in get_sid_by_nthash(construct.ad_id, res[0]).all():
				sidd[sid] = 1

			for sid1 in sidd:
				for sid2 in sidd:
					if sid1 == sid2:
						continue
					self.add_edge(sid1, sid2, construct, label = 'pwsharing')
					cnt += 1

		logger.info('Added %s password sharing edges' % cnt)




from flask import send_file, send_from_directory

def serve_file(filename):
    response = send_from_directory('site/nui/dist', filename)
    response.direct_passthrough = False
    return response

def serve_index():
    response = send_from_directory('site/nui/dist', 'index.html')
    response.direct_passthrough = False
    return response

def serve_favico():
    response = send_from_directory('site/static/img', 'favicon.ico')
    response.direct_passthrough = False
    return response



from . import Basemodel, lf
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey

class JackDawADTrust(Basemodel):
	__tablename__ = 'trusts'
	
	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer, ForeignKey('ads.id'))
	ad = relationship("JackDawADInfo", back_populates="trusts", lazy = True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	cn = Column(String, index=True)
	dn = Column(String, index=True)
	guid = Column(String, index=True)
	
	name = Column(String, index=True)	
	whenChanged = Column(DateTime)	
	whenCreated = Column(DateTime)
	securityIdentifier = Column(String, index=True)
	trustDirection = Column(Integer, index = True)
	trustPartner = Column(String, index=True)
	trustPosixOffset = Column(Integer, index=True)
	trustType = Column(Integer, index = True)
	trustAttributes = Column(String, index=True)
	flatName = Column(String, index=True)

	@staticmethod
	def from_ldapdict(d):
		trust = JackDawADTrust()
		trust.cn = lf(d.get('cn'))
		trust.dn = lf(d.get('distinguishedName'))
		trust.guid = lf(d.get('objectGUID'))

		trust.name = lf(d.get('name'))
		trust.securityIdentifier = str(lf(d.get('securityIdentifier')))
		trust.whenChanged = lf(d.get('whenChanged'))
		trust.whenCreated = lf(d.get('whenCreated'))

		t = d.get('trustDirection')
		trust.trustDirection = t.name if t is not None else None
		trust.trustPartner = lf(d.get('trustPartner'))
		trust.trustPosixOffset = lf(d.get('trustPosixOffset'))
		t = d.get('trustType')
		trust.trustType = t.name if t is not None else None
		trust.trustAttributes = lf(d.get('trustAttributes'))
		trust.flatName = lf(d.get('flatName'))

		return trust


	def to_dict(self):
		return {
			'id' : self.id ,
			'ad_id' : self.ad_id ,
			'cn' : self.cn ,
			'dn' : self.dn ,
			'guid' : self.guid ,
			'name' : self.name ,
			'securityIdentifier' : self.securityIdentifier ,
			'whenChanged' : self.whenChanged ,
			'whenCreated' : self.whenCreated ,
			'trustDirection' : self.trustDirection ,
			'trustPartner' : self.trustPartner ,
			'trustPosixOffset' : self.trustPosixOffset ,
			'trustType' : self.trustType ,
			'trustAttributes' : self.trustAttributes ,
			'flatName' : self.flatName ,
		}


from . import Basemodel
import datetime
from sqlalchemy import Column, Integer, String, DateTime, Boolean

class SMBFinger(Basemodel):
	__tablename__ = 'smbfinger'
	
	id = Column(Integer, primary_key=True)
	fetched_at = Column(DateTime, default=datetime.datetime.utcnow)
	machine_id = Column(Integer)
	signing_enabled = Column(Boolean, index=True)
	signing_required = Column(Boolean, index=True)
	domainname = Column(String, index=True)
	computername = Column(String, index=True)
	dnscomputername = Column(String, index=True)
	dnsdomainname = Column(String, index=True)
	local_time = Column(DateTime, index=True)
	dnsforestname = Column(String, index=True)
	os_major_version = Column(String, index=True)
	os_minor_version = Column(String, index=True)
	os_build = Column(String, index=True)
	os_guess = Column(String, index=True)
	
	@staticmethod
	def from_extra_info(machine_id, extra_info):
		f = SMBFinger()
		f.machine_id = machine_id
		f.signing_enabled = extra_info['signing_enabled']
		f.signing_required = extra_info['signing_required']
		if 'ntlm_data' in extra_info:
			f.domainname = extra_info['ntlm_data']['domainname']
			f.computername = extra_info['ntlm_data']['computername']
			f.dnscomputername = extra_info['ntlm_data']['dnscomputername']
			f.dnsdomainname = extra_info['ntlm_data']['dnsdomainname']
			f.local_time = extra_info['ntlm_data']['local_time']
			f.dnsforestname = extra_info['ntlm_data']['dnsforestname']
			f.os_major_version = extra_info['ntlm_data']['os_major_version']
			f.os_minor_version = extra_info['ntlm_data']['os_minor_version']
			f.os_build = extra_info['ntlm_data']['os_build']
			f.os_guess = extra_info['ntlm_data']['os_guess']

		return f




import math

class GraphNode:
	def __init__(self, gid, name, domainid, gtype = None, properties = {}):
		self.name = name
		self.id = gid
		self.domainid = domainid
		self.type = gtype
		self.properties = properties
		self.mindistance = math.inf

	def set_distance(self, d):
		self.mindistance = min(self.mindistance, d)

	def serialize_mindistance(self):
		if self.mindistance == math.inf:
			return 999999

	def to_dict(self, format = None):
		if format is None:
			return {
				'id' : self.id,
				'name' : self.name,
				'domainid' : self.domainid,
				'properties' : self.properties,
				'md' : self.serialize_mindistance(),
			}

		elif format == 'd3':
			return {
				'id' : self.id,
				'name' : self.name,
				'domainid' : self.domainid,
				'type' : self.type,
				'md' : self.serialize_mindistance(),
			}
		
		elif format == 'vis':
			return {
				'id' : self.id,
				'label' : self.name,
				'type' : self.type,
				'domainid' : self.domainid,
				'md' : self.serialize_mindistance(),
			}

class GraphEdge:
	def __init__(self, src, dst, label = '', weight = 1, properties = {}):
		self.src = src
		self.dst = dst
		self.label = label
		self.weight = weight
		self.properties = properties

	def to_dict(self, format = None):
		if format is None:
			return {
				'src' : self.src,
				'dst' : self.dst,
				'label' : self.label,
				'weight' : self.weight,
				'properties' : self.properties
			}
		elif format == 'd3':
			return {
				'source' : self.src,
				'target' : self.dst,
				'label'  : self.label,
				'weight' : self.weight,
			}
		elif format == 'vis':
			return {
				'from' : self.src,
				'to' : self.dst,
				'label'  : self.label,
				'weight' : self.weight,
			}


class GraphData:
	def __init__(self):
		self.nodes = {}
		self.edges = {}

	def add_node(self, gid, name, domainid, node_type, properties = {}):
		self.nodes[gid] = GraphNode(gid, name, domainid, node_type, properties)
	
	def add_edge(self, src, dst, label = '', weight = 1, properties = {}):
		if src not in self.nodes:
			raise Exception('Node with id %s is not present' % src)
		if dst not in self.nodes:
			raise Exception('Node with id %s is not present' % dst)
		
		key = str(src) + str(dst) + str(label)

		self.edges[key] = GraphEdge(src, dst, label, weight, properties)

	def __add__(self, o):
		if not isinstance(o, GraphData):
			raise Exception('Cannot add GraphData and %s' % type(o))
		
		self.nodes.update(o.nodes)
		self.edges.update(o.edges)
		return self

	def to_dict(self, format = None):
		if format is None:
			return {
				'nodes' : [self.nodes[x].to_dict() for x in self.nodes],
				'edges' : [self.edges[x].to_dict() for x in self.edges]
			}
		elif format == 'd3':
			return {
				'nodes' : [self.nodes[x].to_dict(format = format) for x in self.nodes],
				'links' : [self.edges[x].to_dict(format = format) for x in self.edges]
			}
		elif format == 'vis':
			return {
				'nodes' : [self.nodes[x].to_dict(format = format) for x in self.nodes],
				'edges' : [self.edges[x].to_dict(format = format) for x in self.edges]
			}


from . import Basemodel, lf
import datetime
import hashlib
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, ForeignKey, Index, DateTime, Boolean

from pypykatz.pypykatz import pypykatz


class LSASecret(Basemodel):
	__tablename__ = 'lsasecret'

	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer)
	machine_id = Column(Integer, nullable= False, default = -1)
	secret_type = Column(String, index=True, nullable = False)
	secret = Column(String, index=False, nullable = False)

	def __init__(self, secret_type = None, secret = None, ad_id = -1, machine_id = -1):
		self.ad_id = ad_id
		self.machine_id = machine_id
		self.secret_type = secret_type
		self.secret = secret

	@staticmethod
	def from_cached_secrets(cached_secrets, ad_id = -1, machine_id = -1):
		for secret in cached_secrets:
			s = LSASecret()
			s.ad_id = ad_id
			s.machine_id = machine_id
			s.secret_type = str(type(secret))
			s.secret = str(secret)
			yield s


from . import Basemodel, lf
import datetime
import hashlib
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, ForeignKey, Index, DateTime

from pypykatz.pypykatz import pypykatz
from pypykatz.utils.crypto.winhash import LM, NT

# It may be tempting to use SIDs to link credentials with users in the domain
# However some credentials format don't give SIDs (impacket) others have SIDs 
# that only identify the primary user, but not the owner of the actual credential
# Summary: we need to use the ad_id and the username<->samaccountname
#

class Credential(Basemodel):
	__tablename__ = 'credentials'
	__table_args__ = (Index('Credential_uc', "ad_id", "domain", "username","nt_hash", "lm_hash", "history_no", unique = True), )

	id = Column(Integer, primary_key=True)
	ad_id = Column(Integer)
	machine_id = Column(Integer, nullable= False, default = -1)
	domain = Column(String, index=True, nullable= False)
	username = Column(String, index=True, nullable= False)
	nt_hash = Column(String, index=True, nullable= False)
	lm_hash = Column(String, index=True, nullable= False)
	krb_des_cbc = Column(String, index=True, nullable= True)
	krb_aes128 = Column(String, index=True, nullable= True)
	krb_aes256 = Column(String, index=True, nullable= True)
	krb_rc4_hmac = Column(String, index=True, nullable= True)
	history_no = Column(Integer, index=True, nullable= False)
	cred_type = Column(String, index=True, nullable= False)
	object_sid = Column(String, index=True, nullable= False)
	object_rid = Column(String, index=True, nullable= False)
	pwd_last_set = Column(DateTime, index = True)
	
	def __init__(self, domain = None, username = None, nt_hash = None, lm_hash = None, history_no = None, ad_id = -1):
		self.domain = domain
		self.username = username
		self.nt_hash = nt_hash
		self.lm_hash = lm_hash
		self.history_no = history_no
		self.ad_id = ad_id

	@staticmethod
	def from_samsecret(samsecret, ad_id = -1, machine_id = -1):
		cred = Credential()
		cred.ad_id = ad_id
		cred.machine_id = machine_id
		cred.domain = 'LOCAL'
		cred.username = samsecret.username
		cred.nt_hash = samsecret.nt_hash.hex() if samsecret.nt_hash is not None else None
		cred.lm_hash = samsecret.lm_hash.hex() if samsecret.lm_hash is not None else None
		cred.history_no = 0
		cred.rid = samsecret.rid
		cred.cred_type = 'pypykatz-registry-sam'

		return cred

	@staticmethod
	def from_aiosmb_secret(secret, ad_id = -1):
		# returns a complex touple in the format of (currentcred, [nthist], [lmhist], [cleartextcred], [pwds])
		creds = []
		pwds = []
		
		cred = Credential()
		cred.ad_id = ad_id
		cred.domain = secret.domain
		cred.username = secret.username
		cred.nt_hash = secret.nt_hash.hex() if secret.nt_hash is not None else None
		cred.lm_hash = secret.lm_hash.hex() if secret.lm_hash is not None else None
		cred.cred_type = 'aiosmb-dcsync-ntlm'
		cred.pwd_last_set = secret.pwd_last_set
		cred.history_no = 0

		if secret.object_sid is not None:
			cred.object_sid = str(secret.object_sid)
			t = str(secret.object_sid)
			_, rid = t.rsplit('-',1)
			cred.rid = rid

		for ktype, key in secret.kerberos_keys:
			if str(ktype) == 'aes128-cts-hmac-sha1-96':
				cred.krb_aes128 = str(key)
			elif str(ktype) == 'aes256-cts-hmac-sha1-96':
				cred.krb_aes256 = str(key)
			elif str(ktype) == 'des-cbc-md5':
				cred.krb_des_cbc = str(key)
			elif str(ktype) == 'rc4_hmac':
				cred.krb_rc4_hmac = str(key)

		creds.append(cred) # this is the main one
		
		if secret.cleartext is not None:
			for pw in secret.cleartext:
				cred = Credential()
				cred.ad_id = ad_id
				cred.domain = secret.domain
				cred.username = secret.username
				
				cred.nt_hash = NT(str(pw)).hex()
				cred.lm_hash = None
				cred.history_no = 0
				cred.cred_type = 'aiosmb-dcsync-cleartext'

				creds.append(cred)
				pwds.append(str(pw))

		if secret.lm_history is not None:
			for i, lm in enumerate(secret.lm_history):
				cred = Credential()
				cred.ad_id = ad_id
				cred.domain = secret.domain
				cred.username = secret.username
				cred.lm_hash = lm.hex()
				cred.history_no = i + 1
				cred.cred_type = 'aiosmb-dcsync-ntlm-history'

				creds.append(cred)

		if secret.nt_history is not None:
			for i, nt in enumerate(secret.nt_history):
				cred = Credential()
				cred.ad_id = ad_id
				cred.domain = secret.domain
				cred.username = secret.username
				cred.nt_hash = nt.hex()
				cred.history_no = i + 1
				cred.cred_type = 'aiosmb-dcsync-ntlm-history'

				creds.append(cred)
		
		return creds, pwds


	@staticmethod
	def from_impacket_line(line, ad_id = -1):
		cred = Credential()
		userdomainhist, flags, lm_hash, nt_hash, *t = line.split(':')
		#parsing history
		m = userdomainhist.find('_history')
		history_no = 0
		if m != -1:
			history_no = int(userdomainhist.split('_history')[1]) + 1
			userdomainhist = userdomainhist.split('_history')[0]
		m = userdomainhist.find('\\')
		domain = '<LOCAL>'
		username = userdomainhist
		if m != -1:
			domain = userdomainhist.split('\\')[0]
			username = userdomainhist.split('\\')[1]
		cred.ad_id = ad_id
		cred.nt_hash = nt_hash
		cred.lm_hash = lm_hash
		cred.history_no = history_no
		cred.username = username
		cred.domain = domain
		cred.cred_type = 'dcsync'
		return cred

	@staticmethod
	def from_impacket_stream(stream, ad_id = -1):
		for line in stream:
			yield Credential.from_impacket_line(line.decode(), ad_id = ad_id)

	@staticmethod
	def from_impacket_file(filename, ad_id = -1):
		"""
		Remember that this doesnt populate the foreign keys!!! You'll have to do it separately!
		important: historyno will start at 0. This means all history numbers in the file will be incremented by one
		"""
		with open(filename, 'r') as f:
			for line in f:
				yield Credential.from_impacket_line(line, ad_id = ad_id)

	@staticmethod
	def from_lsass_stream(stream, ad_id = -1):
		from pypykatz.pypykatz import pypykatz
		mimi = pypykatz.parse_minidump_buffer(stream)
		return Credential.lsass_generator(mimi, ad_id = ad_id)

	@staticmethod
	def from_lsass_dump(filename, ad_id = -1):
		
		mimi = pypykatz.parse_minidump_file(filename)
		return Credential.lsass_generator(mimi, ad_id = ad_id)

	@staticmethod
	def lsass_generator(mimi, ad_id):

		for luid in mimi.logon_sessions:
			sid = mimi.logon_sessions[luid].sid

			for cred in mimi.logon_sessions[luid].msv_creds:
				cr = Credential()
				cr.ad_id = ad_id
				cr.nt_hash = cred.NThash.hex() if cred.NThash is not None else '31d6cfe0d16ae931b73c59d7e0c089c0'
				cr.lm_hash = cred.LMHash if cred.LMHash is not None else 'aad3b435b51404eeaad3b435b51404ee'
				cr.history_no = 0
				cr.username = cred.username if cred.username is not None else 'NA'
				cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
				cr.cred_type = 'msv'
				yield cr, None, sid

			for cred in mimi.logon_sessions[luid].wdigest_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'wdigest'
					yield cr, cred.password, sid

			for cred in mimi.logon_sessions[luid].ssp_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'ssp'
					yield cr, cred.password, sid

			for cred in mimi.logon_sessions[luid].livessp_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'live_ssp'
					yield cr, cred.password, sid

			#for cred in mimi.logon_sessions[luid]['dpapi_creds']:
			# dpapi credentials are not used in this database (for now)

			for cred in mimi.logon_sessions[luid].kerberos_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'kerberos'
					yield cr, cred.password, sid

			for cred in mimi.logon_sessions[luid].credman_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'credman'
					yield cr, cred.password, sid

			for cred in mimi.logon_sessions[luid].tspkg_creds:
				if cred.password is not None:
					cr = Credential()
					cr.ad_id = ad_id
					cr.nt_hash = NT(cred.password).hex()
					cr.lm_hash = None
					cr.history_no = 0
					cr.username = cred.username if cred.username is not None else 'NA'
					cr.domain = cred.domainname if cred.domainname is not None else '<LOCAL>'
					cr.cred_type = 'tpskg'
					yield cr, cred.password, sid

				
	@staticmethod
	def from_aiosmb_stream(stream, ad_id = -1):
		for line in stream:
			if line is None or len(line) == 0:
				continue
			yield Credential.from_aiosmb_line(line.decode(), ad_id = ad_id)

	@staticmethod
	def from_aiosmb_file(filename, ad_id = -1):
		"""
		Remember that this doesnt populate the foreign keys!!! You'll have to do it separately!
		important: historyno will start at 0. This means all history numbers in the file will be incremented by one
		"""
		with open(filename, 'r') as f:
			for line in f:
				if line is None or len(line) == 0:
					continue
				yield Credential.from_aiosmb_line(line, ad_id = ad_id)


	@staticmethod
	def from_aiosmb_line(line, ad_id = -1):
		cred = Credential()
		line = line.strip()
		data = line.split(':')
		pw = None
		if data[0] == 'ntlm':
			cred.ad_id = ad_id
			cred.domain = data[1]
			cred.username = data[2]
			cred.nt_hash = data[5]
			cred.lm_hash = data[4]
			cred.history_no = 0
			cred.cred_type = 'aiosmb-dcsync-ntlm'

		elif data[0] == 'ntlm_history':
			cred.ad_id = ad_id
			cred.domain = data[1]
			cred.username = data[2]
			cred.nt_hash = data[5]
			cred.lm_hash = data[4]
			cred.history_no = int(data[6].replace('history_',''))
			cred.cred_type = 'aiosmb-dcsync-ntlm-history'

		elif data[0] == 'cleartext':
			cred.ad_id = ad_id
			_, cred.domain, cred.username, pw = line.split(':', 4) #reparsing needed, pw might contain colon
			
			cred.nt_hash = NT(pw).hex()
			cred.lm_hash = None
			cred.history_no = 0
			cred.cred_type = 'aiosmb-dcsync-cleartext'

		return cred, pw



from flask import current_app
from jackdaw.dbmodel.aduser import JackDawADUser
from jackdaw.dbmodel.adcomp import JackDawADMachine
from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.smbfinger import SMBFinger

def list_machines(domainid, page, maxcnt):
	db = current_app.db
	res = {
		'res' : [],
		'page': {},
	}
	qry = db.session.query(
		JackDawADMachine
		).filter_by(ad_id = domainid
		).with_entities(
			JackDawADMachine.id, 
			JackDawADMachine.objectSid, 
			JackDawADMachine.sAMAccountName
		)
	
	qry = qry.paginate(page = page, max_per_page = maxcnt)

	domains = []
	for uid, sid, name in qry.items:
		domains.append(
			{
				'id' : uid, 
				'sid' : sid, 
				'name': name
			}
		)

	page = dict(
		total=qry.total, 
		current_page=qry.page,
		per_page=qry.per_page
	)

	res['res'] = domains
	res['page'] = page

	return res

def get(domainid, machineid):
	db = current_app.db
	machine = db.session.query(JackDawADMachine).get(machineid)
	return machine.to_dict()

def get_sid(domainid, sid):
	db = current_app.db
	for machine in db.session.query(JackDawADMachine).filter_by(objectSid = sid).filter(JackDawADMachine.ad_id == domainid).all():
		return machine.to_dict()

def get_os_versions(domainid):
	db = current_app.db
	qry = db.session.query(
		JackDawADMachine.operatingSystemVersion
		).filter_by(ad_id = domainid
		).group_by(JackDawADMachine.operatingSystemVersion
		).distinct(JackDawADMachine.operatingSystemVersion)

	versions = {}
	for version in qry.all():
		versions[version[0]] = 1
	
	return list(versions.keys())

def get_domains(domainid):
	db = current_app.db
	qry = db.session.query(
			SMBFinger.domainname
			).filter(JackDawADMachine.ad_id == domainid
			).filter(JackDawADInfo.id == domainid
			).filter(SMBFinger.machine_id == JackDawADMachine.id
			).group_by(SMBFinger.domainname)

	domains = {}
	for domain in qry.all():
		domains[domain[0]] = 1
	
	return list(domains.keys())


from setuptools import setup, find_packages
import re

VERSIONFILE="jackdaw/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))


setup(
	# Application name:
	name="jackdaw",

	# Version number (initial):
	version=verstr,

	# Application author details:
	author="Tamas Jos",
	author_email="skelsecprojects@gmail.com",

	# Packages
	packages=find_packages(),

	# Include additional files into the package
	include_package_data=True,
	
	# Details
	url="https://github.com/skelsec/jackdaw",

	zip_safe = False,
	#
	# license="LICENSE.txt",
	description="Gathering shiny things from your domain",

	# long_description=open("README.txt").read(),
	python_requires='>=3.6',
	classifiers=(
		"Programming Language :: Python :: 3.6",
		"Operating System :: OS Independent",
	),
	install_requires=[
		'aiosmb>=0.2.13',
		'msldap>=0.2.10',
		'sqlalchemy',
		'dnspython',
		'tqdm',
		'networkx',
		'connexion',
		'flask-sqlalchemy',
		'connexion[swagger-ui]',
		'pypykatz>=0.3.6',
        'swagger-ui-bundle>=0.0.2',
		'werkzeug==0.16.1'
	],
	entry_points={
		'console_scripts': [
			'jackdaw = jackdaw.__main__:main',
		],
	}
)



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import asyncio
import logging
import json
import traceback
import ipaddress
import multiprocessing
import threading

from tqdm import tqdm
from dns import resolver, reversename

import aiosmb
from aiosmb.commons.interfaces.machine import SMBMachine
from aiosmb.commons.utils.extb import format_exc

from jackdaw.common.apq import AsyncProcessQueue
from jackdaw.dbmodel.netshare import NetShare
from jackdaw.dbmodel.netsession import NetSession
from jackdaw.dbmodel.localgroup import LocalGroup
from jackdaw.dbmodel.smbfinger import SMBFinger
from jackdaw import logger
from jackdaw.dbmodel import get_session

from jackdaw.dbmodel.adinfo import JackDawADInfo
from jackdaw.dbmodel.adcomp import JackDawADMachine

class SMBEnumeratorProgress:
	def __init__(self):
		self.type = 'SMB'
		self.msg_type = 'PROGRESS'
		self.adid = None
		self.domain_name = None
		self.errors = None
		self.sessions = None
		self.shares = None
		self.groups = None
		self.hosts = None

	def __str__(self):
		if self.msg_type == 'PROGRESS':
			return '[%s][%s][%s][%s] HOSTS %s SHARES %s SESSIONS %s GROUPS %s ERRORS %s' % (
				self.type, 
				self.domain_name, 
				self.adid,
				self.msg_type,
				self.hosts, 
				self.shares, 
				self.sessions, 
				self.groups, 
				self.errors
			
			)
		
		return '[%s][%s][%s][%s]' % (self.type, self.domain_name, self.adid, self.msg_type)

class SMBGathererManager:
	def __init__(self, smb_mgr, worker_cnt = 50, queue_size = 100000, progress_queue = None):
		self.queue_size = queue_size
		self.in_q = None
		self.out_q = None
		self.smb_mgr = smb_mgr
		self.gathering_type = ['all']
		self.localgroups = ['Administrators', 'Distributed COM Users','Remote Desktop Users']
		self.concurrent_connections = worker_cnt if worker_cnt is not None else multiprocessing.cpu_count()
		self.db_conn = None
		self.progress_queue = progress_queue

		self.total_targets = 0
		self.targets = []
		self.targets_file = None
		self.ldap_conn = None
		self.target_ad = None
		self.lookup_ad = None #if specified, it will look up the targets in the DB.
		self.out_file = None

		self.gatherer = None
		self.gatherer_task = None
		self.job_generator_task = None
		self.domain_name = None
		
		self.prg_hosts = None
		self.prg_shares = None
		self.prg_sessions = None
		self.prg_groups = None
		self.prg_errors = None

		self.prg_errors_cnt = 0
		self.prg_sessions_cnt = 0
		self.prg_shares_cnt = 0
		self.prg_groups_cnt = 0
		self.prg_hosts_cnt = 0

		self.results_thread = None

	def __target_generator(self):
		if self.db_conn is not None:
			session = get_session(self.db_conn)
		
		for target in self.targets:
			tid = -1
			yield (tid, target)

		if self.targets_file is not None:
			tid = -1
			with open(self.targets_file, 'r') as f:
				for line in f:
					line = line.strip()
					yield (tid, line)

		if self.ldap_conn is not None:
			ldap_filter = r'(&(sAMAccountType=805306369))'
			attributes = ['sAMAccountName']
			for entry in self.ldap_conn.pagedsearch(ldap_filter, attributes):
				tid = -1
				if self.lookup_ad is not None:
					res = session.query(JackDawADMachine)\
							.filter_by(ad_id = self.lookup_ad)\
							.with_entities(JackDawADMachine.id)\
							.filter(JackDawADMachine.sAMAccountName == entry['attributes']['sAMAccountName'])\
							.first()
					if res is not None:
						tid = res[0]
				
				yield (tid, entry['attributes']['sAMAccountName'][:-1])

		if self.target_ad is not None:
			info = session.query(JackDawADInfo).get(self.target_ad)
			info.smb_enumeration_state = 'STARTED'
			self.domain_name = str(info.distinguishedName).replace(',','.').replace('DC=','')
			session.commit()
			for target_id, target_name in session.query(JackDawADMachine).filter_by(ad_id = self.target_ad).with_entities(JackDawADMachine.id, JackDawADMachine.sAMAccountName):
				yield (target_id, target_name[:-1])

		if self.db_conn is not None:
			session.close()

	async def terminate(self):
		if self.job_generator_task is not None:
			self.job_generator_task.cancel()
		
		if self.gatherer_task is not None:
			await self.gatherer.terminate()
			self.gatherer_task.cancel()
			
	async def generate_targets(self):
		for target in self.__target_generator():
			self.total_targets += 1
			if self.progress_queue is None:
				self.prg_hosts.total = self.total_targets
			await self.in_q.put(target)
		
		await self.in_q.put(None)
	
	async def run(self):
		logger.info('[+] Starting SMB information acqusition. This might take a while...')
		self.in_q = asyncio.Queue(self.queue_size)
		self.out_q = asyncio.Queue(self.queue_size)
		if self.progress_queue is None:
			self.prg_hosts = tqdm(desc='HOSTS', ascii = True)
			self.prg_shares = tqdm(desc='Shares', ascii = True)
			self.prg_sessions = tqdm(desc='Sessions', ascii = True)
			self.prg_groups = tqdm(desc='LocalGroup', ascii = True)
			self.prg_errors = tqdm(desc='Errors', ascii = True)
		
		else:
			msg = SMBEnumeratorProgress()
			msg.msg_type = 'STARTED'
			msg.adid = self.target_ad
			msg.domain_name = self.domain_name
			await self.progress_queue.put(msg)

		#self.results_thread = threading.Thread(target = self.get_results)
		#self.results_thread.daemon = True
		#self.results_thread.start()

		self.gatherer = AIOSMBGatherer(
			self.in_q, 
			self.out_q, 
			self.smb_mgr, 
			gather = self.gathering_type, 
			localgroups = self.localgroups, 
			concurrent_connections = self.concurrent_connections,
			progress_queue = self.progress_queue
		)
		self.gatherer_task = asyncio.create_task(self.gatherer.run())
		
		self.job_generator_task = asyncio.create_task(self.generate_targets())
		
		session = None
		if self.db_conn is not None:
			session = get_session(self.db_conn)
		
		while True:
			x = await self.out_q.get()
			if x is None:
				break

			tid, target, result, error = x
			if result is None and error is not None:
				#something went error
				logger.debug('[AIOSMBScanner][TargetError][%s] %s' % (target.get_ip(), error))
				if self.progress_queue is None:
					self.prg_errors.update()
				else:
					self.prg_errors_cnt += 1

			if result is not None:
				if self.progress_queue is None:
					if isinstance(result, NetSession):
						self.prg_sessions.update()
					elif isinstance(result, NetShare):
						self.prg_shares.update()
					elif isinstance(result, LocalGroup):
						self.prg_groups.update()
				
				else:
					if isinstance(result, NetSession):
						self.prg_sessions_cnt += 1
					elif isinstance(result, NetShare):
						self.prg_shares_cnt += 1
					elif isinstance(result, LocalGroup):
						self.prg_groups_cnt += 1

				if session is None:
					logger.debug(target, str(result), error)
				else:
					session.add(result)
					session.commit()

			if result is None and error is None:
				logger.debug('Finished: %s' % target.ip)
				if self.progress_queue is None:
					self.prg_hosts.update()
				else:
					self.prg_hosts_cnt += 1
					msg = SMBEnumeratorProgress()
					msg.adid = self.target_ad
					msg.domain_name = self.domain_name
					msg.errors = self.prg_errors_cnt
					msg.sessions = self.prg_sessions_cnt
					msg.shares = self.prg_shares_cnt
					msg.groups = self.prg_groups_cnt
					msg.hosts = self.prg_hosts_cnt

					await self.progress_queue.put(msg)

		
		logger.info('[+] SMB information acqusition finished!')
		if self.progress_queue is not None:
			msg = SMBEnumeratorProgress()
			msg.msg_type = 'FINISHED'
			msg.adid = self.target_ad
			msg.domain_name = self.domain_name
			await self.progress_queue.put(msg)

		if session is not None and self.target_ad is not None:
			info = session.query(JackDawADInfo).get(self.target_ad)
			info.smb_enumeration_state = 'FINISHED'
			session.commit()


class AIOSMBGatherer:
	def __init__(self, in_q, out_q, smb_mgr, gather = ['all'], localgroups = [], concurrent_connections = 10, progress_queue = None):
		#multiprocessing.Process.__init__(self)
		self.in_q = in_q
		self.out_q = out_q
		self.smb_mgr = smb_mgr
		self.gather = gather
		self.localgroups = localgroups
		self.concurrent_connections = concurrent_connections
		self.progress_queue = progress_queue

		self.worker_tasks = []
		self.targets = []
		self.worker_q = None

	async def scan_host(self, atarget):
		try:
			tid, target = atarget
			#spneg = AuthenticatorBuilder.to_spnego_cred(self.credential, target)
			connection = self.smb_mgr.create_connection_newtarget(target)
			async with connection:
				await connection.login()
				
				extra_info = connection.get_extra_info()
				if extra_info is not None:
					try:
						f = SMBFinger.from_extra_info(tid, extra_info)
						await self.out_q.put((tid, connection.target, f, None))
					except:
						traceback.print_exc()

				machine = SMBMachine(connection)


				if 'all' in self.gather or 'shares' in self.gather:
					async for smbshare, err in machine.list_shares():
						if err is not None:
							await self.out_q.put((tid, connection.target, None, 'Failed to list shares. Reason: %s' % format_exc(err)))
							continue
						share = NetShare()
						share.machine_id = tid
						share.ip = connection.target.get_ip()
						share.netname = smbshare.name
						share.type = smbshare.type
						share.remark = smbshare.remark

						await self.out_q.put((tid, connection.target, share, None))
					
				
				if 'all' in self.gather or 'sessions' in self.gather:
					async for session, err in machine.list_sessions():
						if err is not None:
							await self.out_q.put((tid, connection.target, None, 'Failed to get sessions. Reason: %s' % format_exc(err)))
							continue

						sess = NetSession()
						sess.machine_id = tid
						sess.source = connection.target.get_ip()
						sess.ip = session.ip_addr.replace('\\','').strip()
						sess.username = session.username

						await self.out_q.put((tid, connection.target, sess, None))

				if 'all' in self.gather or 'localgroups' in self.gather:
					for group_name in self.localgroups:
						async for domain_name, user_name, sid, err in machine.list_group_members('Builtin', group_name):
							if err is not None:
								await self.out_q.put((tid, connection.target, None, 'Failed to connect to poll group memeberships. Reason: %s' % format_exc(err)))
								continue

							lg = LocalGroup()
							lg.machine_id = tid
							lg.ip = connection.target.get_ip()
							lg.hostname = connection.target.get_hostname()
							lg.sid = sid
							lg.groupname = group_name
							lg.domain = domain_name
							lg.username = user_name
							await self.out_q.put((tid, connection.target, lg, None))
		
		except asyncio.CancelledError:
			return

		except Exception as e:
			await self.out_q.put((tid, connection.target, None, 'Failed to connect to host. Reason: %s' % format_exc(e)))
			return

		finally:
			await self.out_q.put((tid, connection.target, None, None)) #target finished

	async def worker(self):
		while True:
			try:
				target = await self.worker_q.get()
				if target is None:
					return
				try:
					await self.scan_host(target)
				except:
					#exception should be handled in scan_host
					continue
			except asyncio.CancelledError:
				return
			except Exception as e:
				logger.exception('WORKER ERROR')
				return

	async def terminate(self):
		for worker in self.worker_tasks:
			worker.cancel()

	async def run(self):
		"""
		Reads targets from queue and scans them
		"""
		try:
			self.worker_q = asyncio.Queue()
			
			for _ in range(self.concurrent_connections):
				self.worker_tasks.append(asyncio.create_task(self.worker()))

			while True:
				target = await self.in_q.get()
				if target is None:
					for _ in range(self.concurrent_connections):
						await self.worker_q.put(None)
					break
				else:
					await self.worker_q.put(target)

			results = await asyncio.gather(*self.worker_tasks, return_exceptions = True)
			for res in results:
				if isinstance(res, Exception):
					logger.error('Error! %s' % res)
			await self.out_q.put(None)
		except:
			import traceback
			traceback.print_exc()






from jackdaw.dbmodel import create_db, get_session, Credential, HashEntry, JackDawADUser
from sqlalchemy import exc, func
from jackdaw import logger
import string

class JackDawCredentials:
	def __init__(self, db_conn, domain_id = -1, db_session = None, cracking_speed = 60 * 10**9):
		self.domain_id = domain_id
		self.db_conn = db_conn
		self.dbsession = db_session

		self.cracking_speed = cracking_speed #default = 10x 1080TI #password/second

	def get_cracking_time(self, he):
		alphabet_size = 0

		if he.pw_lower is True:
			alphabet_size += len(string.ascii_lowercase)

		if he.pw_upper is True:
			alphabet_size += len(string.ascii_uppercase)

		if he.pw_digit is True:
			alphabet_size += 10

		if he.pw_special is True:
			alphabet_size += len(" !\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~ ") #from hashcat

		res = (alphabet_size**he.pw_length) // self.cracking_speed
		return res

	def get_dbsession(self):
		if not self.dbsession:
			self.dbsession = get_session(self.db_conn)

	def add_credentials_impacket(self, impacket_file):
		self.get_dbsession()
		ctr = 0
		ctr_fail = 0
		try:
			for cred in Credential.from_impacket_file(impacket_file, self.domain_id):
				try:
					self.dbsession.add(cred)
					if ctr % 10000 == 0:
						logger.info(ctr)
						self.dbsession.commit()
					
				except exc.IntegrityError as e:
					ctr_fail += 1
					self.dbsession.rollback()
					continue
				else:
					ctr += 1

			self.dbsession.commit()
			
			logger.info('Added %d users. Failed inserts: %d' % (ctr, ctr_fail))
		except Exception as e:
			logger.exception()
		finally:
			self.dbsession.close()

	def add_cracked_passwords_gen(self, gen, disable_usercheck, disable_passwordcheck):
		mem_filter = {}
		self.get_dbsession()
		try:
			ctr = 0
			for he in gen:
				if he.nt_hash in mem_filter:
					continue
				mem_filter[he.nt_hash] = 1

				exists = False

				if disable_passwordcheck is False:
					#check if hash is already in HashEntry, if yes, skip
					if he.nt_hash:
						exists = self.dbsession.query(HashEntry.id).filter_by(nt_hash=he.nt_hash).scalar() is not None
					elif he.lm_hash:
						exists = self.dbsession.query(HashEntry.id).filter_by(lm_hash=he.lm_hash).scalar() is not None
					else:
						continue

					#print(exists)
					if exists is True:
						continue
				
				
				if disable_usercheck is False:
					#check if hash actually belongs to a user, if not, skip, otherwise put it in DB
					if he.nt_hash:
						qry = self.dbsession.query(Credential.nt_hash).filter(Credential.nt_hash == he.nt_hash)
					elif he.lm_hash:
						qry = self.dbsession.query(Credential.lm_hash).filter(Credential.lm_hash == he.lm_hash)
					else:
						continue
					
					exists = True if qry.first() else False
				
				if exists is True:
					try:
						self.dbsession.add(he)
						if ctr % 10000 == 0:
							logger.info(ctr)
						#	self.dbsession.commit()
						self.dbsession.commit()
						

					except exc.IntegrityError as e:
						logger.exception(e)
						self.dbsession.rollback()
						continue
					else:
						ctr += 1

			self.dbsession.commit()
			logger.info('Added %d plaintext passwords to the DB' % ctr)

		except Exception as e:
			logger.exception(e)
		finally:
			self.dbsession.close()
	
	def add_cracked_passwords(self, potfile, disable_usercheck, disable_passwordcheck):
		gen = HashEntry.from_potfile(potfile)
		self.add_cracked_passwords_gen(gen, disable_usercheck, disable_passwordcheck)
		
	def get_uncracked_hashes(self, hash_type, history):
		self.get_dbsession()
		try:
			if hash_type == 'NT':
				qry = self.dbsession.query(Credential.nt_hash).outerjoin(HashEntry, Credential.nt_hash == HashEntry.nt_hash).filter(Credential.nt_hash != None).distinct(Credential.nt_hash)
			else:
				qry = self.dbsession.query(Credential.lm_hash).outerjoin(HashEntry, Credential.lm_hash == HashEntry.lm_hash).filter(Credential.lm_hash != None).distinct(Credential.lm_hash)
			
			if history == False:
				qry = qry.filter(Credential.history_no == 0)
				
			for some_hash in qry.all():
				yield some_hash[0]
		except Exception as e:
			print(e)
		finally:
			self.dbsession.close()

	def get_cracked_users(self):
		qry = self.dbsession.query(JackDawADUser, Credential, HashEntry
						).outerjoin(HashEntry, Credential.nt_hash == HashEntry.nt_hash
						).filter(HashEntry.nt_hash != None
						).filter(Credential.history_no == 0
						).filter(JackDawADUser.ad_id == self.domain_id
						).filter(Credential.ad_id == self.domain_id
						).filter(JackDawADUser.sAMAccountName == Credential.username
						)

		cracked_users = []
		cracked_users.append(['domain', 'username', 'canLogon', 'hashtype', 'pw_length', 'pw_lower', 'pw_upper', 'pw_digit', 'pw_special', 'crack_times_sec'])
		for user, cred, he in qry.all():
			res = [str(cred.domain), str(user.sAMAccountName), str(user.canLogon), 'NT', str(he.pw_length), str(he.pw_lower), str(he.pw_upper), str(he.pw_digit), str(he.pw_special), str(self.get_cracking_time(he))]
			cracked_users.append(res)
		
		return cracked_users

	def get_pwsharing(self):
		occurrence_ctr = {}
		pw_sharing_cracked = 0
		pw_sharing_notcracked = 0
		pw_sharing_total = 0
		pw_sharing_hash = {}

		qry = self.dbsession.query(
				func.count(Credential.nt_hash),
					Credential.nt_hash
					).filter(Credential.history_no == 0
					).filter(Credential.ad_id == self.domain_id
					).group_by(
						Credential.nt_hash
					).having(
						func.count(Credential.nt_hash) > 1
					)

		for occurrence, nthash in qry.all():
			pw_sharing_total += 1
			pw_sharing_hash[nthash] = []
			if occurrence not in occurrence_ctr:
				occurrence_ctr[occurrence] = 0
			occurrence_ctr[occurrence] += 1

			users = []
			for user in self.dbsession.query(Credential.username
									).filter(Credential.history_no == 0
									).filter(Credential.ad_id == self.domain_id
									).filter_by(nt_hash = nthash
									).distinct(Credential.username):
				users.append(user.username)
					
			pw_sharing_hash[nthash] += users
		
		for nthash in pw_sharing_hash:
			exists = self.dbsession.query(HashEntry.id).filter_by(nt_hash=nthash).scalar() is not None
			if exists is True:
				pw_sharing_cracked += 1
			else:
				pw_sharing_notcracked += 1

		#anonymizing (removing nt hash from key, replacing with an int)
		new_pwshare = {}
		for i, key in enumerate(pw_sharing_hash):
			new_pwshare[i] = pw_sharing_hash[key]
		

		return pw_sharing_total, pw_sharing_cracked, pw_sharing_notcracked, new_pwshare

	def cracked_stats(self):
		
		pw_sharing_total, pw_sharing_cracked, pw_sharing_notcracked, _ = self.get_pwsharing()

		total_objects = self.dbsession.query(Credential
			).filter(Credential.history_no == 0
			).filter(Credential.history_no == 0
			).filter(Credential.ad_id == self.domain_id
			).count()
		
		
		#how many users
		total_comps = self.dbsession.query(Credential
			).filter(Credential.username.endswith('$')
			).filter(Credential.history_no == 0
			).filter(Credential.ad_id == self.domain_id
			).count()
			
		#how many users
		total_users = total_objects - total_comps

			
		#how many total cracked
		total_objects_cracked = self.dbsession.query(Credential, HashEntry
			).outerjoin(HashEntry, Credential.nt_hash == HashEntry.nt_hash
			).filter(HashEntry.nt_hash != None
			).filter(Credential.history_no == 0
			).filter(Credential.ad_id == self.domain_id
			).count()

			
		#how many total cracked
		total_comps_cracked = self.dbsession.query(Credential, HashEntry
			).outerjoin(HashEntry, Credential.nt_hash == HashEntry.nt_hash
			).filter(Credential.username.endswith('$')
			).filter(HashEntry.nt_hash != None
			).filter(Credential.history_no == 0
			).filter(Credential.ad_id == self.domain_id
			).count()

			
		#how many users
		total_users_cracked = total_objects_cracked - total_comps_cracked



		res = {
			'total_objects' : total_objects, 
			'total_comps' : total_comps,
			'total_users' : total_users, 
			'total_objects_cracked' : total_objects_cracked, 
			'total_comps_cracked' : total_comps_cracked, 
			'total_users_cracked' : total_users_cracked, 
			'pw_sharing_total' : pw_sharing_total, 
			'pw_sharing_cracked' : pw_sharing_cracked, 
			'pw_sharing_notcracked' : pw_sharing_notcracked,
		}
		return res

	
	def get_cracked_info(self):
		raise NotImplementedError()




from . import Basemodel, lf, dt, bc
import datetime
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, Boolean

class JackDawADGplink(Basemodel):
	__tablename__ = 'adgplink'

	id = Column(Integer, primary_key=True)	
	ent_id = Column(Integer, index=True)
	gpo_dn = Column(String, index=True)
	order = Column(Integer, index=True)
	
	def to_dict(self):
		return {
			'id' : self.id ,
			'ent_id' : self.name ,
			'gpo_dn' : self.dn ,
			'order' : self.path,
		}
	



import connexion

from sqlalchemy.exc import IntegrityError
from jackdaw.dbmodel.credential import Credential
from jackdaw.dbmodel.hashentry import HashEntry
from jackdaw.dbmodel.netsession import NetSession
from jackdaw.dbmodel.adcomp import JackDawADMachine
from jackdaw.dbmodel.aduser import JackDawADUser
from jackdaw.credentials.credentials import JackDawCredentials
from flask import current_app

from pypykatz.utils.crypto.winhash import LM, NT

def impacket_upload(domainid):
	db = current_app.db
	file_to_upload = connexion.request.files['file_to_upload']
	#print(file_to_upload.read())
	ctr = 0
	fail = 0
	for cred in Credential.from_impacket_stream(file_to_upload.stream, domainid):
		try:
			print(cred)
			db.session.add(cred)
			db.session.commit()
			ctr += 1
		except IntegrityError:
			db.session.rollback()
			fail += 1

	return {'new' : ctr, 'duplicates' : fail }

def lsass_upload(domainid, computername = None):
	db = current_app.db
	file_to_upload = connexion.request.files['file_to_upload']
	#print(file_to_upload.read())
	ctr = 0
	fail = 0
	ctr_plain = 0
	fail_plain = 0
	for cred, plaintext, sid in Credential.from_lsass_stream(file_to_upload.stream, domainid):
		try:
			db.session.add(cred)
			db.session.commit()
			ctr += 1
		except IntegrityError:
			db.session.rollback()
			fail += 1

		if plaintext is not None and len(plaintext) > 0:
			he = HashEntry(plaintext, nt_hash = cred.nt_hash)
			try:
				db.session.add(he)
				db.session.commit()
				ctr_plain += 1
			except IntegrityError:
				db.session.rollback()
				fail_plain += 1

		if computername is not None:

			cname = computername
			if computername[-1] != '$':
				cname = computername + '$'
			comp = db.session.query(JackDawADMachine).filter_by(ad_id = domainid).filter(JackDawADMachine.sAMAccountName == cname).first()
			#print('COMP %s' % comp)
			if comp is None:
				continue
			user = db.session.query(JackDawADUser.sAMAccountName).filter_by(ad_id = domainid).filter(JackDawADUser.objectSid == sid).first()
			#print('USER %s' % user)
			#print('SID %s' % sid )
			if user is None:
				continue

			sess = NetSession()
			sess.machine_id = comp.id
			sess.source = comp.sAMAccountName
			sess.username = user.sAMAccountName
			try:
				db.session.add(sess)
				db.session.commit()
			except IntegrityError:
				db.session.rollback()


	return {'new' : ctr, 'duplicates' : fail, 'pwnew' : ctr_plain, 'pwduplicates' :  fail_plain }

def aiosmb_upload(domainid):
	db = current_app.db
	file_to_upload = connexion.request.files['file_to_upload']
	#print(file_to_upload.read())
	ctr = 0
	fail = 0
	ctr_plain = 0
	fail_plain = 0
	for cred, plaintext in Credential.from_aiosmb_stream(file_to_upload.stream, domainid):
		try:
			db.session.add(cred)
			db.session.commit()
			ctr += 1
		except IntegrityError:
			db.session.rollback()
			fail += 1

		if plaintext is not None and len(plaintext) > 0:
			he = HashEntry(plaintext, nt_hash = cred.nt_hash)
			try:
				db.session.add(he)
				db.session.commit()
				ctr_plain += 1
			except IntegrityError:
				db.session.rollback()
				fail_plain += 1

	return {'new' : ctr, 'duplicates' : fail, 'pwnew' : ctr_plain, 'pwduplicates' :  fail_plain }

def potfile_upload():
	disable_usercheck = False
	disable_passwordcheck = False
	file_to_upload = connexion.request.files['file_to_upload']

	db = current_app.db
	creds = JackDawCredentials(None, db_session = db.session)
	gen = HashEntry.from_potfile_stream(file_to_upload.stream)

	creds.add_cracked_passwords_gen(gen, disable_usercheck, disable_passwordcheck)
	

	return {}

def passwords_upload(passwords):
	def pwit(passwords):
		for pw in passwords:
			nt_hash = NT(pw).hex()
			print(pw)
			print(nt_hash)
			yield HashEntry(pw, nt_hash=nt_hash)
	
	db = current_app.db
	disable_usercheck = False
	disable_passwordcheck = False

	gen = pwit(passwords)
	creds = JackDawCredentials(None, db_session = db.session)
	creds.add_cracked_passwords_gen(gen, disable_usercheck, disable_passwordcheck)

	return {}

def passwords_upload_file():
	def pwit(fs):
		for line in fs:
			line = line.decode()
			line = line.strip()

			nt_hash = NT(line).hex()
			yield HashEntry(line, nt_hash=nt_hash)
	
	file_to_upload = connexion.request.files['file_to_upload']
	db = current_app.db
	disable_usercheck = False
	disable_passwordcheck = False

	gen = pwit(file_to_upload.stream)
	creds = JackDawCredentials(None, db_session = db.session)
	creds.add_cracked_passwords_gen(gen, disable_usercheck, disable_passwordcheck)

	return {}

def get_uncracked_current(domainid, hashtype = 'nt'):
	hashtype = hashtype.upper()
	db = current_app.db
	creds = JackDawCredentials(None, domain_id = domainid, db_session = db.session)
	hashes = []
	for data in creds.get_uncracked_hashes(hashtype, False):
		hashes.append(data)
	return hashes, 200

def get_uncracked_all(domainid, hashtype = 'nt'):
	hashtype = hashtype.upper()
	db = current_app.db
	creds = JackDawCredentials(None, domain_id = domainid, db_session = db.session)
	hashes = []
	for data in creds.get_uncracked_hashes(hashtype, True):
		hashes.append(data)
	return hashes, 200

def get_cracked_users(domainid):
	db = current_app.db
	creds = JackDawCredentials(None, domain_id = domainid, db_session = db.session)
	rows = creds.get_cracked_users()
	return rows, 200

def get_pwsharing(domainid):
	db = current_app.db
	creds = JackDawCredentials(None, domain_id = domainid, db_session = db.session)
	pw_sharing_total, pw_sharing_cracked, pw_sharing_notcracked, new_pwshare = creds.get_pwsharing()
	return {
		'pw_sharing_total' : pw_sharing_total, 
		'pw_sharing_cracked' : pw_sharing_cracked,
		'pw_sharing_notcracked' : pw_sharing_notcracked,
		'pwsharing_users' : new_pwshare
	}

def get_stats(domainid):
	db = current_app.db
	creds = JackDawCredentials(None, domain_id = domainid, db_session = db.session)
	res = creds.cracked_stats()
	return res





import connexion
from flask import current_app

from aiosmb.commons.connection.url import SMBConnectionURL
from msldap.commons.url import MSLDAPURLDecoder
from jackdaw.gatherer.smb.smb import SMBGathererManager
from jackdaw.gatherer.ldap.aioldap import LDAPEnumeratorManager


def scan_enum(params):
	db_conn = current_app.config['SQLALCHEMY_DATABASE_URI']

	ldap_url = params['ldap_url']
	smb_url = params['smb_url']
	ldap_workers = params['ldap_workers']
	smb_workers = params['smb_workers']

	smb_mgr = SMBConnectionURL(smb_url)
	ldap_mgr = MSLDAPURLDecoder(ldap_url)

	mgr = LDAPEnumeratorManager(db_conn, ldap_mgr, agent_cnt=ldap_workers)
	adifo_id = mgr.run()
	#print('ADInfo entry successfully created with ID %s' % adifo_id)
		
	mgr = SMBGathererManager(smb_mgr, worker_cnt=smb_workers)
	mgr.gathering_type = ['all']
	mgr.db_conn = db_conn
	mgr.target_ad = adifo_id
	mgr.run()

	return {'adifo_id' : adifo_id}
